# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
from typing import cast
from typing import Any
from typing import ClassVar
from typing import Callable

import pydantic

from canonical.lib.utils.encoding import b64decode_json
from canonical.ext.jose.protocols import ITokenSignatureVerifier
from canonical.ext.jose.types import JWSCompactEncoded
from canonical.ext.jose.types import JWTDict
from .jwt import JWT
from .signature import Signature


class JWS(pydantic.BaseModel):
    CompactEncoded: ClassVar[type[JWSCompactEncoded]] = JWSCompactEncoded
    signatures: list[Signature] = pydantic.Field(
        default=...
    )
    claims: dict[str, Any] = {}
    payload: bytes

    @classmethod
    def require(cls, func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(func)
        def f(self: pydantic.RootModel[Any], *args: Any, **kwargs: Any):
            if isinstance(self.root, cls.CompactEncoded):
                self.root = self.root.jose(cls)
            if not isinstance(self.root, cls):
                raise TypeError(
                    f"{type(self).__name__} is not a JSON Web Signature."
                )
            return func(self, *args, **kwargs)
        return f

    def compact(self):
        return JWSCompactEncoded(bytes.decode(self.encode_compact(), 'ascii'))

    def encode(self, compact: bool = True) -> bytes:
        """Encodes the JWS/JWE."""
        if len(self.signatures) > 1: # pragma: no cover
            raise ValueError("Compact encoding can not be used with multiple signatures.")
        return self.encode_compact()

    def encode_compact(self) -> bytes:
        assert len(self.signatures) == 1
        return self.signatures[0].encode(self.payload)

    def is_jwt(self) -> bool:
        """Return a boolean indicating if any of the signature headers says
        that the payload is a JWT.
        """
        return any([str.lower(h.typ or '') == 'jwt' for h in self.signatures])

    def __str__(self):
        return bytes.decode(bytes(self), 'ascii')

    def __bytes__(self):
        return self.encode_compact()

    async def verify(self, verifier: ITokenSignatureVerifier) -> bool:
        """Return a boolean indicating if at least one signature
        validated using the given verifier.
        """
        claims: JWTDict = {}
        if self.is_jwt():
            # TODO: This might raise an exception if the client specified
            # the typ header as JWT but the payload isn't.
            jwt = JWT.model_validate(b64decode_json(self.payload))
            claims = cast(JWTDict, jwt.model_dump())
        if not self.signatures:
            return False
        for signature in self.signatures:
            is_valid = await signature.verify(
                verifier=verifier,
                claims=claims,
                payload=self.payload
            )
            if not is_valid:
                continue
            break
        else:
            is_valid = False
        return is_valid


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import Literal
from typing import TypeVar

import pydantic

from canonical.ext.oauth.types import AuthenticationMethod
from .requestmodel import RequestModel


G = TypeVar('G')
ClientAssertionType = Literal[
    'urn:ietf:params:oauth:client-assertion-type:jwt-bearer'
]

RFC7523 = 'https://datatracker.ietf.org/doc/html/rfc7523'


class GrantModel(RequestModel, Generic[G]):
    grant_type: G = pydantic.Field(
        default=...,
        title="Grant Type",
        description=(
            "Specifies the grant type requested by the client."
        )
    )

    client_id: str | None = pydantic.Field(
        default=None,
        title="Client ID",
        description=(
            "Required, if the client is not authenticating with "
            "the authorization server or authentication is required "
            "and the client uses the `client_secret_post` method."
        )
    )

    client_secret: str | None = pydantic.Field(
        default=None,
        title="Client Secret",
        description=(
            "Required if the client is authenticating with the "
            "`client_secret_post` method. The client may omit "
            "the parameter if the client secret is an empty "
            "string"
        ),
    )

    client_assertion_type: ClientAssertionType | None = pydantic.Field(
        default=None,
        title="Client Assertion Type",
        description=(
            "The format of the assertion as defined by the authorization "
            "server. The value is an absolute URI. If `client_assertion_"
            "type` is provided, then `client_secret` **must** be omitted."
            
        )
    )

    client_assertion: str | None = pydantic.Field(
        default=None,
        title="Client Assertion",
        description=(
            "The client assertion in the format and encoding indicated "
            "by the `client_assertion_type` field.\n\n"
            "This server supports the following assertion types:\n\n"
            "- `urn:ietf:params:oauth:client-assertion-type:jwt-bearer` - A "
            "JSON Web Token (JWT) signed by a trusted issuer that holds the "
            f"claims identifying the client. Refer to (RFC7523)[{RFC7523}] "
            "for a complete description of the claims and their meanings."
        )
    )

    resource: list[str] = pydantic.Field(
        default_factory=list,
        title="Resources",
        description=(
            "Indicates the target service or protected resource "
            "where the client intends to use the requested access token."
        )
    )

    def assert_client(self, type: ClientAssertionType, assertion: str):
        self.client_id = self.client_secret = None
        self.client_assertion_type = type
        self.client_assertion = assertion

    def identify(self, client_id: str) -> None:
        self.client_id = client_id

    def set_client_secret(self, mode: AuthenticationMethod, secret: str) -> None:
        if mode != 'client_secret_post':
            raise NotImplementedError
        self.client_secret = secret


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Union

import pydantic

from .dependencyreference import DependencyReference


LiteralType = Union[str, int, float, bool, list[Any]]


class DependencyParameter(pydantic.RootModel[DependencyReference | LiteralType]):

    def is_reference(self) -> bool:
        return isinstance(self.root, DependencyReference)

    def resolve(self):
        if self.is_reference():
            raise NotImplementedError
        return self.root


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import datetime
import functools
from typing import cast
from typing import Any
from typing import AsyncIterator
from typing import Generic
from typing import Iterable
from typing import Mapping
from typing import TypeVar

import pydantic
from google.cloud.datastore import Client
from google.cloud.datastore.query import PropertyFilter

from canonical.exceptions import DoesNotExist
from canonical.exceptions import MultipleObjectsReturned
from canonical.ext.resource import IResourceQuery
from .protocols import IDatastoreCursor
from .protocols import IDatastoreQuery
from .protocols import IDatastoreKey
from .protocols import IDatastoreEntity


T = TypeVar('T')


class DatastoreCursor(IResourceQuery[T], Generic[T]):
    _client: Client
    _filters: Iterable[tuple[str, str, int | str | datetime.datetime]]
    _keys_only: bool
    _keys: list[IDatastoreKey]
    _kind: str
    _limit: int | None
    _loop: asyncio.AbstractEventLoop
    _model: type[pydantic.BaseModel]
    _sort: list[str]

    def __init__(
        self,
        kind: str,
        model: type[T],
        client: Client,
        keys: list[IDatastoreKey] | None = None,
        filters: Iterable[tuple[str, str, int | str | datetime.datetime]] | None = None,
        namespace: str | None = None,
        sort: Iterable[str] | None = None,
        page_size: int = 1000,
        limit: int | None = None,
        _keys_only: bool = False
    ):
        self._client = client
        self._filters = filters or []
        self._keys_only = _keys_only
        self._keys = keys or []
        self._kind = kind
        self._limit = limit
        self._loop = asyncio.get_running_loop()
        self._model = cast(type[pydantic.BaseModel], model)
        self._namespace = namespace
        self._page_size = page_size
        self._sort = list(sort or [])

    def factory(self, entity: IDatastoreEntity):
        return cast(T, self._model.model_validate(dict(entity)))

    def model_factory(self, entity: Mapping[str, Any] | IDatastoreEntity):
        return self._model.model_validate(entity)

    def keys(self, page_size: int | None = None, limit: int | None = None) -> 'EntityKeyDatastoreCursor':
        return EntityKeyDatastoreCursor(
            kind=self._kind,
            model=self._model, # type: ignore
            client=self._client,
            keys=self._keys,
            filters=self._filters,
            sort=self._sort,
            page_size=page_size or self._page_size,
            limit=limit or self._limit,
            namespace=self._namespace,
            _keys_only=True
        )

    def get_query(self) -> IDatastoreQuery:
        return cast(
            IDatastoreQuery,
            self._client.query(kind=self._kind, namespace=self._namespace) # type: ignore
        )

    async def all(self) -> AsyncIterator[T]:
        cursor: bytes | None = None
        while True:
            c = await self.run_query(limit=self._page_size, page=cursor)
            objects = list(c)
            if not objects:
                break
            for entity in objects:
                yield self.factory(entity)
            if self._keys:
                # TODO: It is assumed here that with a keys query, all objects
                # are returned in one call.
                break
            if not c.next_page_token:
                break
            cursor = c.next_page_token

    async def exists(self) -> bool:
        return await self.keys().first() is not None

    async def first(self) -> T | None:
        c = await self.run_query(limit=1)
        objects = list(c)
        if not objects:
            return None
        return self.factory(objects[0])

    async def one(self) -> T:
        c = await self.run_query(limit=2)
        objects = list(c)
        if len(objects) > 1:
            raise MultipleObjectsReturned
        if not objects:
            raise DoesNotExist
        return self.factory(objects[0])

    async def run_query(self, limit: int | None = None, page: bytes | None = None) -> IDatastoreCursor:
        if self._keys:
            f = functools.partial(self._client.get_multi, self._keys) # type: ignore
        else:
            q = self.get_query()
            for attname, op, value in self._filters:
                q.add_filter(filter=PropertyFilter(attname, op, value)) # type: ignore
            if self._sort:
                q.order = self._sort
            f = functools.partial(q.fetch, start_cursor=page, limit=limit)
        return await self._loop.run_in_executor(None, f) # type: ignore

    async def __aiter__(self):
        async for obj in self.all():
            yield obj

    def __await__(self):
        return self._iterall().__await__()

    async def _iterall(self):
        return [x async for x in self]


class EntityKeyDatastoreCursor(DatastoreCursor[IDatastoreKey]): # type: ignore

    def factory(self, entity: IDatastoreEntity) -> IDatastoreKey:
        return entity.key

    def get_query(self) -> IDatastoreQuery:
        q = super().get_query()
        q.keys_only()
        return q


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import Self
from typing import TypeVar

import pydantic

from canonical.exceptions import ProgrammingError
from .objectmeta import ObjectMeta
from .resource import M
from .resource import Resource
from .resourcestatus import ResourceStatus


S = TypeVar('S', bound=ResourceStatus)


class StatefulResource(Resource[M], Generic[M, S]):
    _adapter: ClassVar[pydantic.TypeAdapter[Any]]

    status: S = pydantic.Field(
        default=...,
        description=(
            "The `status` field reports the current state of the resource. This "
            "value is modified by the system and can not be changed by clients. "
            "If the `.status` field is `null`, then the resource is created "
            "but no component or system has reported any state yet. Under "
            "normal circumstances, a state is set post-creation, and the "
            "absense of a state usually indicates an error."
        )
    )

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any) -> None:
        super().__pydantic_init_subclass__(**kwargs)
        field = cls.model_fields['status']
        ResourceStatusImpl = field.annotation
        if not inspect.isclass(ResourceStatusImpl):
            return
        if not issubclass(ResourceStatusImpl, ResourceStatus):
            return
        cls._adapter = pydantic.TypeAdapter(ResourceStatusImpl)
        cls._status_model = ResourceStatusImpl
        assert cls.model_rebuild(force=True)

    @classmethod
    def has_state(cls) -> bool:
        return True

    @classmethod
    def model_validate_input(cls, data: dict[str, Any]) -> Self:
        data.setdefault('status', cls._status_model.default())
        return super().model_validate_input(data)

    @property
    def status_adapter(self) -> pydantic.TypeAdapter[S]:
        return self._adapter

    def clusterstate(self):
        assert self.status is not None
        return self.status.clusterstate()

    def initialize_status(self, status: S) -> None:
        pass

    def is_final(self) -> bool:
        return False

    def model_post_init(self, _: Any) -> None:
        super().model_post_init(_)
        self.status.attach(self)

    def _initialize_status(self, force: bool = False):
        self.status.attach(self)
        try:
            self.initialize_status(self.status)
        except pydantic.ValidationError as e:
            raise ProgrammingError(
                f"{type(self).__name__} raised an exception of "
                f"type {type(e).__name__}. Make sure that the "
                f"method {type(self).__name__}.initialize_status() "
                "is properly implemented."
            ) from e

    def _on_replaced(self, old: Self):
        super()._on_replaced(old)
        assert isinstance(self.metadata, ObjectMeta)
        assert isinstance(self.metadata.uid, str) or self.metadata.uid > 0
        self._initialize_status(force=True)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable


NOT_AVAILABLE = object()


class Deferred:

    def __init__(self, resolve: Callable[[], Any]):
        self.__resolve = resolve
        self.__result = NOT_AVAILABLE

    def __str__(self) -> str:
        if self.__result == NOT_AVAILABLE:
            self.__result = self.__resolve()
        return str(self.__result)

    def __getattr__(self, attname: str) -> Any:
        if self.__result == NOT_AVAILABLE:
            self.__result = self.__resolve()
        return getattr(self.__result, attname)


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
import logging
from typing import cast
from typing import Any
from typing import Generic
from typing import TypeVar
from typing import TYPE_CHECKING

import fastapi

from canonical.ext.resource import ResourceType
from ..params import ResourceModel
from ..params import ResourceRepository
from ..params import RequestObjectReference
from ..params import RequestResource
if TYPE_CHECKING:
    from .base import BaseOperation


T = TypeVar('T')


class Validator(Generic[T]):
    """Validates the input data for verbs."""
    __module__: str = 'canonical.ext.fastapi'
    logger: logging.Logger = logging.getLogger('uvicorn')
    model: type[T]
    old: T | None
    new: T | None

    def __init__(
        self,
        request: fastapi.Request,
        repo: ResourceRepository,
        model: ResourceModel,
        key: RequestObjectReference,
        old: RequestResource
    ):
        self.model = cast(type[T], model)
        self.namespace = getattr(request.state, 'namespace', None)
        self.key = key
        self.new = None
        self.old = cast(T | None, old)
        self.repo = repo
        self.request = request

    def get_validator_func(self, verb: 'BaseOperation[Any]'):
        return getattr(self, f'on_{verb.verb}', None)

    async def check_references(
        self,
        repo: ResourceRepository,
        obj: T
    ) -> None:
        return

    async def validate(self, verb: 'BaseOperation[Any]', obj: T | None = None) -> T:
        """Runs the validators for the specified resource."""
        self.fail = verb.fail
        self.new = obj
        self.verb = verb
        if verb.requires_body:
            if self.new is None:
                verb.fail(422, "Input required.")
            if verb.creates and self.old is not None:
                self.fail(409, "Object exists.")
            resource = cast(ResourceType, obj)
            namespace = getattr(resource.metadata, 'namespace', None)
            if namespace != self.namespace:
                verb.fail(403,
                    f"Not allowed to mutate this object in namespace '{namespace}' "
                    f"using endpoint {self.request.url.path} (namespace: {self.namespace})."
                )

            # Handle replace.
            if self.old and self.new:
                _ = cast(ResourceType, self.new)
                old = cast(ResourceType, self.old)

                # The object must be replaceable.
                if not old.replacable() and verb.replaces():
                    self.fail(409, "Object can not be replaced")

            await self.check_references(self.repo, self.new)

        func = self.get_validator_func(verb)
        if func is not None:
            result = func(new=self.new, old=self.old)
            if inspect.isawaitable(result):
                await result
        else:
            self.logger.warning(
                "%s does not define a validation handler for verb '%s'.",
                type(self).__name__,
                verb.verb
            )
        assert self.new is not None
        return self.new


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any
from typing import Callable
from typing import Generic
from typing import Protocol
from typing import TypeVar

import pydantic

from ..transactionparty import BaseTransactionParty


D = TypeVar('D', covariant=True, bound=pydantic.BaseModel)
G = TypeVar('G')
P = TypeVar('P')
R = TypeVar('R')
T = TypeVar('T', bound=BaseTransactionParty[Any, Any], contravariant=True)


class IPartyResolver(Protocol, Generic[P, R, D, T]):
    logger: logging.Logger = logging.getLogger('canonical.ext.ea')

    def descriptor(
        self,
        party: P,
        role: R,
        require: set[str],
        *,
        optional: set[str],
        **kwargs: Any
    ) -> D:
        ...

    def on_created(self, party: P, ref: T) -> None:
        pass

    def on_role_assigned(self, party: P, role: R) -> None:
        pass

    async def assign_role(self, party: P, role_name: str) -> tuple[R, bool]:
        ...

    async def factory(self, ref: T) -> P:
        ...

    async def get(self, ref: T) -> P | None:
        ...

    async def persist_party(self, instance: P) -> None:
        ...

    async def persist_role(self, instance: R) -> None:
        ...

    async def resolve(
        self,
        party: P,
        txp: T,
        name: str,
        *,
        role: R | None,
        required: bool = True,
        force: bool = False
    ):
        ...

    async def resolve_references(
        self,
        party: P,
        ref: T,
        role: R,
        require: set[str],
        optional: set[str],
        force: bool = False
    ) -> None:
        for name in {*require, *optional}:
            await self.resolve(party, ref, name, role=role, required=name not in optional, force=force)

    async def create(self, ref: T, require: set[str], optional: set[str]) -> P:
        party = await self.factory(ref)
        self.on_created(party, ref)
        await self.persist_party(party)
        return party

    async def transaction(
        self,
        ref: T,
        require: set[str] | None = None,
        optional: set[str] | None = None,
        factory: Callable[[dict[str, Any]], Any] | None = None,
        force: bool = False
    ) -> D:
        require = require or set()
        optional = optional or set()
        party = await self.get(ref)
        if party is None:
            party = await self.create(ref, require, optional)
        role, assigned = await self.assign_role(party, ref.role)
        if assigned:
            self.logger.info(
                "Assigned role to party (party: %s, role: %s)",
                str(party), str(role)
            )
            self.on_role_assigned(party, role)
        await self.resolve_references(party, ref, role, require, optional, force=force)
        descriptor = self.descriptor(party, role, require,
            optional=optional, display_name=ref.display_name)
        if factory:
            descriptor = factory(descriptor.model_dump())
        return descriptor


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Generic
from typing import TypeVar


T = TypeVar('T')
EncryptionResult = TypeVar('EncryptionResult')


class IDataEncryption(Generic[EncryptionResult]):

    async def decrypt(
        self,
        ct: EncryptionResult,
        decoder: type[T] = bytes
    ) -> T:
        ...

    async def encrypt(
        self,
        value: Any,
        encoder: type = bytes
    ) -> EncryptionResult:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .icache import ICache
from .ideferrable import IDeferrable
from .imaskable import IMaskable
from .iquery import IQuery
from .irepository import IRepository
from .iresourceidentifier import IResourceIdentifier
from .istorage import IStorage
from .itemplate import ITemplateService
from .itransaction import ITransaction
from .ityped import ITyped


__all__: list[str] = [
    'ICache',
    'IDeferrable',
    'IMaskable',
    'IQuery',
    'IRepository',
    'IResourceIdentifier',
    'IStorage',
    'ITemplateService',
    'ITransaction',
    'ITyped',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import contextlib
import copy
import secrets
from typing import Any
from typing import TypeVar

import httpx
import pydantic

from canonical.ext.cache import MemoryCache
from canonical.ext.httpx import AsyncClient
from canonical.ext.jose.protocols import ITokenSigner
from canonical.lib.protocols import ICache
from canonical.lib.types import HTTPResourceLocator
from canonical.lib.utils.http import MediaTypeSelector
from .models import OIDCProvider
from .models import ServerMetadata
from .models.requests import GrantModel
from .models.responses import TokenResponse
from .protocols import IObtainingClient
from .types import AuthenticationMethod
from .types import Error
from .types import ProtocolViolation
from .types import ResponseTypeLiteral


T = TypeVar('T', bound=pydantic.BaseModel)


class Client(IObtainingClient[TokenResponse]):
    __module__: str = 'canonical.ext.oauth'
    cache_ttl: int = 7200
    mode: AuthenticationMethod | None
    selector = MediaTypeSelector({"application/json"})

    @property
    def id(self) -> str:
        return self.client_id

    def __init__(
        self,
        server: str,
        client_id: str,
        *,
        client_secret: str | None = None,
        signer: ITokenSigner | None = None,
        mode: AuthenticationMethod | None = None,
        token_endpoint: str | None = None,
        authorization_endpoint: str | None = None,
        metadata_url: str | None = None,
        cache: ICache = MemoryCache(scope='global')
    ):
        self.cache = cache
        self.client_id = client_id
        self.client_secret = client_secret
        self.mode = mode
        self.signer = signer
        if mode is None and client_secret:
            self.mode = 'client_secret_post'
        if mode is None and signer:
            self.mode = 'private_key_jwt'
        if signer:
            self.signer = signer.configure(auto_now=True, auto_jti=True)
        adapter = pydantic.TypeAdapter(ServerMetadata | OIDCProvider)
        self.provider = adapter.validate_python({
            'issuer': server,
            'authorization_endpoint': authorization_endpoint,
            'token_endpoint': token_endpoint,
            'metadata_url': metadata_url,
            'protocol': 'oauth2:oidc'
        })

    def authorize(
        self,
        response_type: ResponseTypeLiteral
    ) -> HTTPResourceLocator:
        """Return a string containing the URL a user-agent must redirect
        to in order to authorize a grant.
        """
        assert isinstance(self.provider, ServerMetadata)
        if not self.provider.authorization_endpoint:
            raise NotImplementedError
        params: dict[str, str] = {
            'client_id': self.client_id,
            'state': secrets.token_hex()
        }
        if response_type:
            params['response_type'] = response_type
        return self.provider.authorization_endpoint.with_query(**params)

    def configure(self, **kwargs: Any) -> IObtainingClient[TokenResponse]:
        client = copy.deepcopy(self)
        client.provider = ServerMetadata.model_validate({
            **client.provider.model_dump(),
            **kwargs
        })
        return client

    async def create_client_assertion(
        self,
        provider: ServerMetadata,
        signer: ITokenSigner
    ) -> str:
        """Create a client assertion."""
        assert self.signer
        claims: dict[str, Any] = {
            'aud': provider.token_endpoint,
            'sub': self.client_id,
        }
        return await self.signer.sign_claims(claims, ttl=120)

    async def discover(self, http: httpx.AsyncClient | None) -> ServerMetadata:
        """Return a :class:`~canonical.ext.oauth.models.ServerMetadata`
        instance describing the authorization server and its capabilities.
        """
        assert self.provider
        if not isinstance(self.provider, ServerMetadata):
            provider = await self.cache.get(self.provider.issuer, ServerMetadata)
            if provider is None:
                async with self.http(http=http) as http:
                    provider = await self.provider.discover(http)
                assert isinstance(self.provider, OIDCProvider)
                await self.cache.set(
                    key=self.provider.issuer,
                    value=provider,
                    encoder=ServerMetadata,
                    ttl=self.cache_ttl
                )
            self.provider = provider
        return self.provider

    @contextlib.asynccontextmanager
    async def http(self, http: httpx.AsyncClient | None = None):
        if http is not None:
            yield http
            return
        async with AsyncClient() as http:
            yield http
            return

    async def obtain(
        self,
        grant: GrantModel[Any],
        http: httpx.AsyncClient | None = None
    ) -> TokenResponse | Error | ProtocolViolation:
        assert isinstance(self.provider, ServerMetadata)
        auth = None
        match self.mode:
            case 'client_secret_basic':
                assert self.client_secret
                auth = httpx.BasicAuth(self.client_id, self.client_secret)
            case 'client_secret_post':
                assert self.client_secret
                grant.identify(self.client_id)
                grant.set_client_secret(self.mode, self.client_secret)
            case 'private_key_jwt':
                assert self.signer is not None
                grant.assert_client(
                    type='urn:ietf:params:oauth:client-assertion-type:jwt-bearer',
                    assertion=await self.create_client_assertion(self.provider, self.signer)
                )
            case 'client_secret_jwt':
                # TODO: Should do the same as private_key_jwt, but with a signer
                # that uses the client_secret.
                raise NotImplementedError(self.mode)
            case None:
                grant.identify(self.client_id)
            case _:
                raise NotImplementedError(self.mode)
        return await self._request(
            request=httpx.Request(
                method='POST',
                url=self.provider.token_endpoint,
                headers={'Content-Type': 'application/x-www-form-urlencoded'},
                content=grant.model_dump_urlencoded()
            ),
            response_model=TokenResponse,
            auth=auth,
            http=http
        )

    async def _request(
        self,
        request: httpx.Request,
        response_model: type[T],
        http: httpx.AsyncClient | None = None,
        auth : httpx.Auth | None = None,
        accept: set[str] = {"application/json"}
    ) -> T | Error | ProtocolViolation:
        adapter: pydantic.TypeAdapter[response_model | Error | ProtocolViolation] =\
            pydantic.TypeAdapter(response_model | Error | ProtocolViolation)
        selector = MediaTypeSelector(accept)
        async with self.http(http) as http:
            response = await http.send(request, auth=auth)
            mt = selector.select(response.headers.get('Content-Type'))
            result = ProtocolViolation.model_validate(
                "The authorization server responded with an unexpected "
                f"content type: {response.headers.get('Content-Type')}."
            )
            if mt == "application/json":
                result = adapter.validate_json(response.content)
        return result

    def __repr__(self):
        return f'<Client: {self.client_id}@{self.provider.issuer}>'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol


class IRegisteredClient(Protocol):

    @property
    def id(self) -> str:
        ...

    def allows_redirect(self, uri: str | None) -> bool:
        """Return a boolean indicating if the client allows
        redirection to the given URI.
        """
        ...

    def allows_response_type(self, response_type: str) -> bool:
        """Return a boolean indicatinf if the client allows the
        given response type.
        """
        ...

    def default_redirect(self) -> str:
        """Return the default redirection URI."""
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import inspect
from typing import Any
from typing import Callable

import fastapi


class Endpoint:
    _is_coroutine = asyncio.coroutines._is_coroutine # type: ignore

    @property
    def __name__(self) -> str:
        return self.name

    def __init__(
        self,
        name: str,
        signature: inspect.Signature,
        handle: Callable[..., Any],
    ):
        self.handle = handle
        self.name = name
        self.__signature__ = signature
        # Check if the asyncio.iscoroutinefunction() call returns
        # True for this object, since it depends on a private
        # symbol.
        assert asyncio.iscoroutinefunction(self) # nosec

    async def __call__(self, *args: Any, **kwargs: Any) -> fastapi.Response:
        response = await self.handle(*args, **kwargs)
        if inspect.isawaitable(response):
            response = await response
        assert isinstance(response, fastapi.Response)
        return response


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .authenticationmethod import AuthenticationMethod
from .error import Error
from .protocolviolation import ProtocolViolation
from .redirecturi import RedirectURI
from .responsetype import ResponseTypeLiteral
from .stopsnooping import StopSnooping
from .tokentype import TokenTypeLiteral
from .urlsafeclientid import URLSafeClientID
from .useragentexception import UserAgentException


__all__: list[str] = [
    'AuthenticationMethod',
    'Error',
    'ProtocolViolation',
    'RedirectURI',
    'ResponseTypeLiteral',
    'StopSnooping',
    'TokenTypeLiteral',
    'URLSafeClientID',
    'UserAgentException',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import AsyncContextManager
from typing import Mapping

import fastapi
import httpx


class ProxiedResponse(fastapi.responses.StreamingResponse):

    def __init__(
        self,
        stream: AsyncContextManager[httpx.Response],
        headers: Mapping[str, str] | None = None,
    ):
        self.stream = stream
        fastapi.Response.__init__(self, content=b'')

    async def stream_response(self, send: Any) -> None:
        async with self.stream as response:
            super().__init__(
                status_code=response.status_code,
                headers={**self.headers, **response.headers},
                content=response.aiter_bytes()
            )
            await send(
                {
                    "type": "http.response.start",
                    "status": response.status_code,
                    "headers": self.raw_headers,
                }
            )
            async for chunk in self.body_iterator:
                if not isinstance(chunk, bytes):
                    chunk = chunk.encode(self.charset)
                await send({"type": "http.response.body", "body": chunk, "more_body": True})

            await send({"type": "http.response.body", "body": b"", "more_body": False})


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from canonical.ext.oauth.bases import AuthorizationServerEndpoint
from canonical.ext.oauth.params import AuthorizationResponse
from canonical.ext.oauth.protocols import IRegisteredClient
from canonical.ext.oauth.resources import AuthorizationState
from canonical.ext.oauth.types import StopSnooping
from .useragentroutehandler import UserAgentRouteHandler


class ClientRedirectionEndpoint(AuthorizationServerEndpoint):
    cache_ttl: int = 3600
    name = 'oauth2.callback'
    params: AuthorizationResponse
    path = '/callback'
    route_class = UserAgentRouteHandler
    response_model = None
    status_code = 302
    summary = 'Redirection Endpoint'

    async def get(self):
        if not self.state.is_forwarding():
            raise NotImplementedError
        assert self.params.code
        self.state.status.update(
            metadata=self.state.metadata,
            status='Granted',
            code=self.params.code
        )
        await self.resources.persist(AuthorizationState, new=self.state)
        return fastapi.responses.RedirectResponse(
            status_code=302,
            url=self.state.get_return_url(
                iss=f'{self.request.url.scheme}://{self.request.url.netloc}'
            )
        )

    async def get_client(self) -> IRegisteredClient | None:
        return None

    async def setup(self):
        if self.params.state is None:
            raise StopSnooping
        try:
            self.state = await self.resources.get(
                key=self.params.state,
                model=AuthorizationState,
                require=True
            )
        except self.resources.DoesNotExist as e:
            raise StopSnooping from e


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal
from typing import Generic
from typing import TypeVar

import pydantic


N = TypeVar('N')
K = TypeVar('K')


class BaseSubject(pydantic.BaseModel, Generic[K, N]):
    api_group: Literal['iam.webiam.io'] = pydantic.Field(
        default='iam.webiam.io',
        alias='apiGroup',
        description=(
            "Must be `iam.webiam.io` for `User`, `ServiceAccount` and "
            "`Group` references."
        )
    )

    kind: K = pydantic.Field(
        default=...,
        description=(
            "Kind of object being referenced. Values defined by this API "
            "group are `User`, `Group`, `ServiceAccount` and `Domain`. If the "
            "Authorizer does not recognized the kind value, the Authorizer "
            "must report an error."
        )
    )

    name: N = pydantic.Field(
        default=...,
        description=(
            "Name of the object being referenced."
        )
    )

    namespace: str | None = pydantic.Field(
        default=None,
        description=(
            "Namespace of the referenced object. If the object kind is "
            "non-namespace, such as `User` or `Group`, and this value "
            "is not empty the Authorizer must report an error."
        )
    )

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any) -> None:
        cls.model_config['title'] = cls.__name__
        cls.model_rebuild()

    def model_post_init(self, _: Any) -> None:
        assert isinstance(self.kind, str)
        if self.kind in {'User', 'Group', 'ServiceAccount'} and self.namespace:
            raise ValueError("Namespace must be unset.")


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from canonical.ext.fastapi.mixins import QueryModelMixin
from canonical.lib import FormDataModel

from .annotations import AuthorizationCode
from .annotations import Issuer
from .annotations import State


class AuthorizationResponse(FormDataModel, QueryModelMixin):
    model_config = {'extra': 'forbid'}
    code: AuthorizationCode = None
    iss: Issuer = None
    state: State = None


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
from typing import Any
from typing import Generic
from typing import Literal
from typing import TypeVar
from typing import TYPE_CHECKING
from typing import Union

import pydantic
from pydantic.fields import FieldInfo

from ..types import APIVersion
if TYPE_CHECKING:
    from .. import APIResourceType
    from ..apiversioned import APIVersioned


T = TypeVar('T', bound=Union['APIResourceType', 'APIVersioned'])


class APIVersionedMeta(pydantic.BaseModel, Generic[T]):
    model_config = {'populate_by_name': True}
    _model: type[T] | None = pydantic.PrivateAttr(default=None)

    #api_version: str
    api_group: str = pydantic.Field(alias='group')
    #base_path: str
    kind: str
    namespaced: bool
    plural: str | None
    short: str | None = None
    version: str
    root: bool = False

    @property
    def api_version(self) -> APIVersion:
        v = f'{self.api_group}/{self.version}'\
            if self.api_group\
            else self.version
        return APIVersion(v)

    @property
    def base_path(self) -> str:
        p = f'{self.api_version}'
        if self.namespaced:
            p = f'{p}/namespaces/{{namespace}}'
        return f'{p}/{self.plural}'

    @property
    def cache_prefix(self):
        return f'{self.api_group or "core"}:{self.version}:{self.plural}'

    @property
    def model(self) -> type[T]:
        assert self._model
        return self._model

    def cache_key(self, name: str, namespace: str | None = None) -> str:
        if self.namespaced and namespace is None:
            raise TypeError("Namespace can not be None.")
        key = self.cache_prefix
        if namespace:
            key = f'{key}:{namespace}'
        return f'{key}:{name}'

    def contribute_to_class(
        self,
        cls: type[T],
        fields: dict[str, FieldInfo],
        root: tuple[type[Any], ...] | None = None
    ):
        setattr(cls, '__meta__', self)
        self._model = cls
        if fields:
            fields['api_version'].annotation = Literal[f'{self.api_version}']
            fields['api_version'].default = self.api_version
            fields['kind'].annotation = Literal[f'{self.kind}']
            fields['kind'].default = self.kind

    def get_url(self, detail: bool = False, subpath: str | None = None) -> str:
        path = f'{self.base_path}'
        if detail:
            path = f'{path}/{{name}}'
        path = f'/{path}'
        if subpath:
            path = f'{path}{subpath}'
        return path

    def is_namespaced(self) -> bool:
        return self.namespaced

    def reference(self, name: str, namespace: str | None = None):
        assert self._model is not None
        # TODO
        from ..objectreference import ObjectReference
        self.validate_namespace(namespace)
        ref = ObjectReference(
            api_version=self.api_version,
            kind=self.kind,
            name=name,
            namespace=namespace or ''
        )
        return ref.with_model(self._model)

    def validate_namespace(self, namespace: str | None):
        if not namespace and self.is_namespaced():
            raise TypeError(
                "The namespace parameter must be provided for "
                f"namespaced resources ({self._model.__name__})."
            )


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal

import pydantic

from .apiresourcelist import APIResourceList


class APIGroupVersionList(pydantic.BaseModel):
    group: str = pydantic.Field(
        default=...
    )

    kind: Literal['APIGroupVersionList'] = pydantic.Field(
        default='APIGroupVersionList'
    )

    versions: list[APIResourceList] = pydantic.Field(
        default_factory=list
    )

    index: dict[str, APIResourceList] = pydantic.Field(
        default_factory=dict,
        exclude=True
    )

    def model_post_init(self, _: Any) -> None:
        self.index = {v.groupVersion: v for v in self.versions}

    def add(self, version: APIResourceList) -> None:
        if not self.index.get(version.groupVersion):
            self.index[version.groupVersion] = version
            self.versions.append(version)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
from typing import cast
from typing import Annotated
from typing import Any
from typing import ClassVar
from typing import Literal
from typing import TYPE_CHECKING

import pydantic
from pydantic.fields import FieldInfo

from ..apimodelfieldinfo import APIModelFieldInfo
from ..apimodelinspector import APIModelInspector
from ..bases import BaseResourceDTO
from ..meta import APIVersionedMeta
from .apimodelclass import APIModelClassBuilder
if TYPE_CHECKING:
    from ..apiresource import APIResource
    from ..objectmeta import ObjectMeta


class APIResourceClassBuilder(APIModelClassBuilder['APIResource[Any]']):
    inspector = APIModelInspector()

    @property
    def api_version(self):
        return str.strip(f'{self.group}/{self.version}', '/')

    def __init__(
        self,
        group: str | None = None,
        version: str | None = None,
        plural: str | None = None,
        **kwargs: Any
    ) -> None:
        super().__init__(**kwargs)
        if self.abstract:
            return
        if group is None:
            raise TypeError("The `group` class parameter can not be None.")
        if version is None:
            raise TypeError("The `version` class parameter can not be None.")
        if plural is None:
            raise TypeError("The `plural` class parameter can not be None.")
        self.group = group
        self.models = self.inspector.get_models(self.model)
        self.plural = plural
        self.version = version

    def construct(self) -> None:
        if self.abstract:
            return
        self.set_field_annotation('api_version', Literal[f'{self.api_version}'])
        self.set_field_default('api_version', self.api_version)
        self.set_field_annotation('kind', Literal[f'{self.model.__name__}'])
        self.set_field_default('kind', self.model.__name__)
        if self.has_metadata():
            self.build_metadata()
        self.rebuild()
        super().construct()

    def setup_namespaced(self):
        """Make a resource namespaced."""
        pass

    def build_metadata(self) -> None:
        field, metadata_class = self.get_metadata_class()
        if self.namespaced:
            metadata_class = metadata_class.with_namespace()
            models = self.inspector.get_models(metadata_class) # type: ignore
            field.annotation = getattr(models, self.building)
            assert metadata_class.model_fields.get('namespace')
        self.metadata_class = self.model.__metadata_class__ = metadata_class

        # Build a descriptor model for the metadata to store it separately
        # from the model. This model should include the api_version
        # and kind fields.
        self.metadata_class.__stored_model__ = type('ObjectDescriptor', (self.metadata_class,), {
            'namespace': '',
            '__annotations__': {
                'api_version': Annotated[str, pydantic.Field()],
                'kind': Annotated[str, pydantic.Field()],
                'namespace': Annotated[str, pydantic.Field()],
            }
        })

    def get_metadata_class(self):
        field = self.fields['metadata']
        return field, cast('type[ObjectMeta[Any]]', field.annotation)

    def get_presentation_bases(self) -> tuple[type[Any]]:
        return BaseResourceDTO,

    def get_subclass_attrs(
        self,
        model: type[pydantic.BaseModel],
        mode: str,
        annotations: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        return {
            **super().get_subclass_attrs(model, mode, annotations={
                **(annotations or {}),
                '__meta__': ClassVar[APIVersionedMeta[Any]]
            }),
            '__meta__': self.model.__meta__,
            'get_object_ref': self.model.get_object_ref,
            'key': cast('APIResource[Any]', model).key,
            'model': property(lambda _: self.model),
            'name': self.model.name,
        }

    def must_include_field(
        self,
        field: FieldInfo | APIModelFieldInfo,
        mode: str
    ):
        return not isinstance(field, APIModelFieldInfo) or field.is_included(mode)

    def on_representation_added(self, mode: str, model: type[APIResource[Any]]):
        model.model_fields['metadata'].annotation = self.subclass(
            model=self.metadata_class,
            mode=mode,
            attrs={},
            bases=(pydantic.BaseModel,)
        )
        assert model.model_rebuild(force=True)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import inspect
from typing import get_args
from typing import get_origin
from typing import Annotated
from typing import Any
from typing import Callable

import fastapi
import fastapi.params

from canonical.ext.cache import NullCache
from canonical.lib.protocols import ICache
from canonical.lib.utils import merge_signatures
from canonical.lib.utils import merged_call


class EndpointRequestHandler:
    _is_coroutine = asyncio.coroutines._is_coroutine # type: ignore

    @property
    def __name__(self) -> str:
        return self.endpoint.__name__

    def __init__(
        self,
        endpoint: type,
        handle: Callable[..., Any],
        cache: ICache = NullCache(),
        **kwargs: Any
    ):
        self.cache = cache
        self.endpoint = endpoint
        self.handle = handle
        self.kwargs = kwargs

        # Check if the asyncio.iscoroutinefunction() call returns
        # True for this object, since it depends on a private
        # symbol.
        assert asyncio.iscoroutinefunction(self) # nosec

        # Inject some parameters that are used by EndpointRequestHandler
        # and Endpoint.
        annotations: dict[str, Any] = {}
        for base in reversed(inspect.getmro(endpoint)):
            annotations.update(inspect.get_annotations(base))
        params: list[inspect.Parameter] = [
            inspect.Parameter(
                kind=inspect.Parameter.POSITIONAL_ONLY,
                name='self_request',
                annotation=fastapi.Request
            )
        ]
        for attname, annotation in annotations.items():
            if get_origin(annotation != Annotated):
                continue
            if len(get_args(annotation)) != 2:
                continue
            annotation, default = get_args(annotation)
            if not isinstance(default, fastapi.params.Depends):
                continue
            params.append(
                inspect.Parameter(
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    name=f'self_{attname}',
                    annotation=annotation,
                    default=default
                )
            )

        # Merge the signatures of the class and the handle method
        # so that they can both receive dependencies.
        self.__signature__ = merge_signatures([
            inspect.signature(endpoint),
            inspect.signature(handle),
        ], extra=params)

    async def __call__(
        self,
        *args: Any,
        **kwargs: Any
    ) -> fastapi.Response:
        attrs = {
            str.lstrip(k, 'self_'): kwargs.pop(k) for k in list(kwargs.keys())
            if str.startswith(k, 'self_')
        }
        endpoint = merged_call(self.endpoint, kwargs)
        endpoint.cache = self.cache
        for attname, value in attrs.items():
            setattr(endpoint, attname, value)
        await endpoint._setup()
        response = await merged_call(self.handle, {**kwargs, 'self': endpoint})
        if response is None:
            response = fastapi.Response(status_code=204)
        return response


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from canonical import DomainName
from .bases import BaseNamespace
from .hierarchicalnamespacespec import HierarchicalNamespaceSpec
from .refs import DomainParentReference


class DomainSpec(HierarchicalNamespaceSpec[DomainParentReference]):
    pass


class Domain(
    BaseNamespace[DomainName],
    group='',
    version='v1',
    plural='domains',
    type='webiam.io/domain'
):
    model_config = {'populate_by_name': True}

    spec: DomainSpec = pydantic.Field(
        description="Defines the behavior of the `Namespace`."
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import NotRequired
from typing import TypedDict

import pydantic

from .base import BaseOperation as Verb
from .command import Command
from .create import Create
from .default import Default
from .defaultcreate import DefaultCreate
from .defaultreplace import DefaultReplace
from .exists import Exists
from .get import Get
from .proxy import Proxy
from .replace import Replace
from .retrieve import Retrieve
from .validator import Validator


__all__: list[str] = [
    'Command',
    'Create',
    'Default',
    'DefaultCreate',
    'DefaultReplace',
    'Exists',
    'Get',
    'Proxy',
    'Replace',
    'Retrieve',
    'Validator',
    'Verb'
]


class VerbParameters(TypedDict):
    detail: NotRequired[bool]
    description: NotRequired[str]
    method: NotRequired[str]
    path: NotRequired[str]
    permissions: NotRequired[set[str]]
    response_model: NotRequired[type[pydantic.BaseModel]]
    status_code: NotRequired[int]
    summary: NotRequired[str]
    validator: NotRequired[type[Validator[Any]]]
    authorize: NotRequired[bool]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
from typing import Any

import fastapi
import fastapi.params

from canonical.exceptions import ProgrammingError
from canonical.ext.resource import IResourceRepository
from .proxiedresponse import ProxiedResponse
from .resourceapplication import ResourceApplication
from .resourceclient import ResourceClient
from .resourcerouter import ResourceRouter
from .utils import inject
from .utils import requires


__all__: list[str] = [
    'inject',
    'requires',
    'EnvironmentVariable',
    'ProxiedResponse',
    'ResourceApplication',
    'ResourceClient',
    'ResourceRouter'
]

def setup_dependencies(
    repository: type[IResourceRepository[Any]]
):
    async def f(
        request: fastapi.Request,
        resources: IResourceRepository[Any] = fastapi.Depends(repository)
    ):
        setattr(request.state, 'resources', resources)

    return [fastapi.Depends(f)]


def EnvironmentVariable(name: str, required: bool = True):
    def f(request: fastapi.Request):
        try:
            return os.environ[name]
        except KeyError:
            if required:
                raise ProgrammingError(
                    f"Dependency {name} is requested from environment "
                    "but it was not present."
                )
            return None

    return fastapi.Depends(f)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal
from typing import TypeVar


T = TypeVar('T')


class ICache:
    __module__: str = 'canonical.lib.protocols'

    async def clear(self, key: str) -> None:
        ...

    async def get(
        self,
        key: str,
        decoder: type[T] = bytes,
        validation: Literal['strict', 'ignore'] = 'strict',
        mode: Literal['json'] = 'json',
        keyalg: Literal['sha256'] | None = None
    ) -> T | None:
        ...

    async def put(
        self,
        key: str,
        value: Any,
        encrypt: bool = False,
        ttl: int | None = None
    ) -> None:
        ...

    async def set(
        self,
        key: str,
        value: Any,
        encoder: type = bytes,
        encrypt: bool = False,
        ttl: int | None = None,
        keyalg: Literal['sha256'] | None = None
    ) -> None:
        """Set a key in the cache. Encode `value` using the given
        encoder if it is not a byte-sequence.
        """
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from canonical.ext.jose import JOSE
from canonical.ext.jose.models import JWS



OPAQUE_JWT = (
    "eyJhbGciOiJSUzI1NiIsImtpZCI6IjkwOWIyZmYwZmYzNjJjMTI0ZjhlMjY0NDZkYW"
    "VjZDU0NGNjZjU0ZDMiLCJ0eXAiOiJKV1QifQ.eyJpc3MiOiAiY29jaGlzZUBtb2xhb"
    "m8tcGkxMW54LXN2Yy5pYW0uZ3NlcnZpY2VhY2NvdW50LmNvbSIsICJzdWIiOiAiY29"
    "tLm1vbGFub2FwaXMuZXhpbSJ9.U0JNNnVRUTFXdEkySXlJVXBGYlgwZ2Z2eUZHMnJn"
    "Z2FoNUFpRjA0dThZTlJRVnI3UnNoZXNHZ3hUaEZGSWRSSWQ1ejBzMWMyeWIzNGJ2S1"
    "RtcHBkR0RYcEhvdC1YWUw2MjJwMXdiVlVJWWVDdG1WMDJ4LUkxZ0FMamoza09rVGR2"
    "N0t2YzNUdlEyaHBfVmVIR002a2w3bFBfSWtmTnpMZk1naGh3dWFZSGphWmhuLWZBOW"
    "g3UG00aFMydlp3cVlKR0lqRFBKSTZ5V0tSOGpqYS03YlN3REs2Wlo0cVJwRGgwWmtv"
    "M1hNV055Q0hSR2RZTm93aWtpbVZwelBfdDNQR2s2NEUzdVZCTXVfVnpWZ2JmdVo1b2"
    "tYWEQ2a2dIYTFkSkRpV2dZU1FfcllGVHhIQ2ZwRjhqR1NaQXJsM2lwOUFDVzVuVDg0"
    "MUYtSkQzakFGNGMwWWtB"
)


def test_encoding_does_not_change_signature():
    o1 = JOSE.model_validate(OPAQUE_JWT)
    o2 = JOSE.model_validate(o1.compact())
    assert isinstance(o1.root, JWS)
    assert isinstance(o2.root, JWS)
    assert o1.compact() == o2.compact()


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from canonical.ext.fastapi.params import HTTPClient
from canonical.ext.fastapi.params import ResourceRepository
from canonical.ext.oauth.models.responses import AuthorizationResponse
from canonical.ext.oauth.resources import AuthorizationState
from canonical.ext.oauth.types import StopSnooping
from canonical.ext.oauth import ClusterClient


async def get(
    http: HTTPClient,
    repo: ResourceRepository,
    params: AuthorizationResponse = AuthorizationResponse.as_query()
):
    if not params.state or not params.code:
        raise NotImplementedError
    try:
        state = await repo.get(params.state, AuthorizationState, require=True)
    except repo.DoesNotExist as e:
        raise StopSnooping from e
    ctrl = state.get_controller(require=True, model=ClusterClient)
    client = await repo.get(ctrl)
    await state.obtain(http, client.spec, params)
    yield state


RedirectedState: TypeAlias = Annotated[
    AuthorizationState,
    fastapi.Depends(get)
]




# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .audiencetype import AudienceType
from .joseheaderdict import JOSEHeaderDict
from .jwasigningalgorithm import JWASigningAlgorithm
from .jwecompactencoded import JWECompactEncoded
from .jwscompactencoded import JWSCompactEncoded
from .jwtdict import JWTDict
from .keyopenum import KeyOpEnum
from .keyuseenum import KeyUseEnum

__all__: list[str] = [
    'AudienceType',
    'JOSEHeaderDict',
    'JWASigningAlgorithm',
    'JWECompactEncoded',
    'JWSCompactEncoded',
    'JWTDict',
    'KeyOpEnum',
    'KeyUseEnum',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import copy
import inspect
import logging
from collections import OrderedDict
from typing import cast
from typing import get_args
from typing import get_origin
from typing import Any
from typing import Annotated
from typing import Callable
from typing import Generic
from typing import NoReturn
from typing import TypeVar

import fastapi
import fastapi.params
import pydantic
from fastapi.exceptions import HTTPException

from canonical.exceptions import Duplicate
from canonical.exceptions import Immutable
from canonical.exceptions import ReferentDoesNotExist
from canonical.exceptions import ProgrammingError
from canonical.ext.api import APIModelInspector
from canonical.ext.api import APIResourceType
from canonical.ext.api.bases import BaseResource
from canonical.ext.api.protocols import IResource
from canonical.ext.cache import NullCache
from canonical.ext.resource import Error
from canonical.ext.resource import ResourceType
from canonical.ext.resource import IResourceRepository
from canonical.ext.iam.protocols import IResourceAuthorizationContext
from canonical.lib.protocols import ICache
from canonical.utils import throw
from .headers import RESPONSE_CONTENT_TYPE
from .headers import RESPONSE_ETAG_RESOURCE
from ..params import AcceptedContentType
from ..params import NegotiateResponseMediaType
from ..params import RequestAuthorizationContext
from ..params import RequestObjectReference
from ..params import RequestResource
from ..params import RequestVerb
from ..params import ResourceRepository
from ..response import Response
from ..utils import inject
from .basemetaclass import BaseMetaclass
from .endpoint import Endpoint
from .validator import Validator


T = TypeVar('T')


class BaseOperation(Generic[T], metaclass=BaseMetaclass):
    authenticated: bool = True
    authorized: bool = True
    cache: ICache = NullCache()
    creates: bool = False
    dependants: list[Callable[..., Any]]
    dependencies: list[Any]
    description: str | None = None
    handle: Any
    include_in_schema: bool = True
    input_model: type[pydantic.BaseModel] | None = None
    input_params: set[str]
    inspector: APIModelInspector = APIModelInspector()
    logger: logging.Logger = logging.getLogger('uvicorn')
    method: str | None = None
    model: type[APIResourceType]
    path: str | None = None
    permissions: set[str]
    summary: str | None = None
    verb: str = ''

    default_responses: dict[int, Any] = {
        401: {
            'model': Error,
            'description': (
                "Authentication is required to perform the requested "
                "operation or the provided credential is invalid."
            ),
        },
        403: {
            'model': Error,
            'description': (
                "Untrusted credential or the authenticated request is not allowed "
                "to perform the requested operation."
            )
        },
        406: {
            'model': Error,
            'description': (
                "The media type accepted by the client can not be "
                "satisfied by the server."
            )
        },
        500: {
            'model': Error,
            'description': (
                "The server encountered an unexpected condition that "
                "prevented it from fulfilling the request."
            )
        }
    }
    detail: bool = False
    existing: bool = False
    requires_body: bool = False
    status_code: int = 200

    def __init__(
        self,
        *,
        model: type[APIResourceType],
        verb: str | None = None,
        method: str | None = None,
        path: str | None = None,
        permissions: set[str] | None = None,
        summary: str | None = None,
        description: str | None = None,
        validator: type[Validator[Any]] = Validator,
        context_class: type[IResourceAuthorizationContext] | None = None,
        response_model: type[pydantic.BaseModel] | None = None,
        authenticated: bool = True,
        detail: bool | None = None,
        status_code: int | None = None,
        ttl: int | None = None,
        authorize: bool = True
    ) -> None:
        self.authorized = authorize
        self.authenticated = authenticated
        self.context_class = context_class
        self.description = description
        if detail is not None:
            self.detail = detail
        if self.detail and not self.creates:
            self.existing = True
        self.input_params = set()
        self.meta = self.inspector.inspect(model)
        self.method = method or self.method
        self.model = model
        self.path = self.path or path
        self.response_model = response_model
        self.status_code = status_code or self.status_code
        self.summary = summary
        self.ttl = ttl
        self.validator_class = validator
        self.views = self.inspector.get_models(model)
        if verb is not None:
            self.verb = verb
        if not self.verb:
            raise TypeError(
                f"{type(self).__name__}.__init__() missing 1 required "
                "positional argument: 'verb'"
            )
        self.dependants = []
        self.dependencies = [
            fastapi.Depends(self.context_class or (lambda: None)),
            fastapi.Depends(self.inject_verb(self.verb)),
            fastapi.Depends(self.inject_model),
        ]
        if not (300 <= self.status_code < 400):
            self.dependencies.extend([
                NegotiateResponseMediaType({
                    'text/html',
                    'application/yaml',
                    'application/json'
                }),
            ])
        self.permissions = set(permissions or [])
        if self.is_namespaced():
            self.dependencies.append(fastapi.Depends(self.inject_namespace))
        if self.is_detail():
            self.dependencies.append(fastapi.Depends(self.inject_name))
        if self.needs_existing():
            self.dependencies.append(fastapi.Depends(self.inject_existing))
        if self.input_model is not None:
            self.requires_body = True
        if self.requires_body:
            self.dependencies.extend([
                AcceptedContentType({'application/yaml', 'application/json'}),
                fastapi.Depends(self.inject_validator())
            ])
        if self.ttl is not None:
            self.dependencies.append(fastapi.Depends(self.inject_cache))

    def add_to_router(
        self,
        router: fastapi.FastAPI | fastapi.APIRouter,
        register: Callable[[type[APIResourceType], str], None],
        authorization: type[IResourceAuthorizationContext] | None = None,
        cache: ICache = NullCache(),
    ):
        self.cache = cache
        if authorization:
            self.dependencies = [
                *inject(resource_context=authorization),
                *self.dependencies
            ]
        router.add_api_route(
            methods=self.get_request_methods(),
            path=self.get_path(),
            endpoint=self.as_handler(),
            name=self.get_endpoint_name(),
            summary=self.get_endpoint_summary(),
            description=getattr(self.handle, '__doc__', None),
            dependencies=[*self.dependencies, fastapi.Depends(self.authorize)],
            include_in_schema=self.include_in_schema,
            responses={**self._get_openapi_responses()},
            response_description=self.get_response_description(),
            response_model=self.get_response_model(),
            response_model_by_alias=True,
            status_code=self.status_code,
            tags=[self.model.__name__],
            operation_id=self.get_endpoint_name()
        )
        register(self.model, self.verb)

    def annotate_handler(
        self,
        signature: inspect.Signature
    ) -> inspect.Signature:
        return signature

    def as_handler(self):
        # Do some modifications on the signatures because for
        # the base implementation classes, not every type is
        # known yet (i.e. the actual model).
        try:
            sig = inspect.signature(self.handle)
        except ValueError:
            raise TypeError(
                "Invalid signature for handler function or method. "
                f"Does {self.handle.__module__}.{self.handle.__name__}() "
                "take a verb.Verb instance as its first positional "
                "argument?"
            )
        params: OrderedDict[str, inspect.Parameter] = OrderedDict(sig.parameters)
        for param in sig.parameters.values():
            if not self.is_injectable(param):
                raise ProgrammingError(
                    f"Parameter '{param.name}' in handler "
                    f"{self.handle.__name__} is not injectable."
                )
            if param.annotation in (ResourceType, IResource, APIResourceType):
                params[param.name] = param.replace(annotation=self.get_model())
                continue
            if self.is_model(param.annotation) and not self.requires_body:
                # In this case, the annotation is assumed to point to the
                # existing resource.
                assert self.needs_existing()
                params[param.name] = param.replace(annotation=RequestResource)
                continue

            params[param.name] = param

        if not params.get('request'):
            params['request'] = inspect.Parameter(
                kind=inspect.Parameter.POSITIONAL_ONLY,
                name='request',
                annotation=fastapi.Request
            )

        sig = sig.replace(
            parameters=list(sorted(params.values(), key=lambda x: x.kind))
        )
        return Endpoint(
            name=self.get_endpoint_name(),
            signature=self.annotate_handler(sig),
            handle=self,
        )

    def depends(self, func: Callable[..., Any]) -> Callable[..., Any]:
        self.dependants.append(func)
        return func

    def fail(self, status_code: int, detail: str) -> NoReturn:
        raise HTTPException(status_code=status_code, detail=detail)

    def get_endpoint_name(self) -> str:
        return f'{self.meta.plural}.{self.verb}'

    def get_endpoint_summary(self) -> str:
        return self.summary or throw(NotImplementedError)

    def get_input_model(self) -> type[pydantic.BaseModel]:
        assert self.views.create is not None
        return self.views.create

    def get_model(self) -> type[pydantic.BaseModel]:
        return self.model

    def get_openapi_responses(self) -> dict[int, dict[str, Any]]:
        responses = copy.deepcopy(self.default_responses)
        if self.detail:
            responses[404] = {
                'model': Error,
                'description': (
                    f'The {self.model.__name__} specified by the path parameter(s) '
                    'does not exist.'
                )
            }
        if self.requires_body:
            responses[415] = {
                'model': Error,
                'description': "Invalid content type for request body."
            }
            responses[422] = {
                'model': Error,
                'description': "Invalid schema for request body."
            }
        if self.method == 'HEAD':
            responses.pop(406)

        if self.is_detail() and self.method in {'PUT', 'PATCH', 'DELETE'}:
            responses[412] = {
                'model': Error,
                'description': (
                    "The `If-Match` header mismatches with "
                    f"the existing {self.model.__name__}."
                )
            }

        return responses

    def get_path(self) -> str:
        path = f'{self.meta.base_path}'
        if self.detail:
            path = f'{path}/{{name}}'
        path = f'/{path}'
        if self.path:
            path = f'{path}{self.path}'
        return path

    def get_request_methods(self) -> list[str]:
        if self.method is None:
            raise ProgrammingError(
                f"{type(self).__name__}.method is None. Specify the attribute "
                f"or override {type(self).__name__}.get_request_methods()."
            )
        return [self.method] 

    def get_response_description(self) -> str:
        return self.description or throw(NotImplementedError)

    def get_response_model(self) -> type[pydantic.BaseModel] | None:
        if self.method == 'HEAD' or not (300 <= self.status_code < 400):
            return None
        if self.response_model:
            return self.response_model
        return self.model if (self.detail or self.creates) else self.model.List # type: ignore

    def has_response_model(self) -> bool:
        return (self.detail or self.creates)

    def inject_cache(
        self,
        cache_control: str | None = fastapi.Header(
            default=None,
            alias="Cache-Control",
            description=(
                "The `Cache-Control` header indicates the client preferences "
                "with regards to caching. Providing the `no-cache` directive "
                "requests the API server to serve to content directly from "
                "the primary data storage."
            )
        )
    ):
        pass

    def inject_name(
        self,
        request: fastapi.Request,
        name: str = fastapi.Path(
            description=f'The `.metadata.name` of an existing resource.',
            max_length=64,
        )
    ):
        setattr(request.state, 'name', name)

    def inject_namespace(
        self,
        request: fastapi.Request,
        namespace: str = fastapi.Path(
            description=(
                "Identifies the namespace that contains the "
                "resource."
            )
        ),
    ):
        setattr(request.state, 'namespace', namespace)

    def inject_model(self, request: fastapi.Request):
        setattr(request.state, 'model', self.model)

    def inject_validator(self):
        def f(request: fastapi.Request, validator: Validator[Any] = fastapi.Depends(self.validator_class)):
            setattr(request.state, 'validator', validator)
        return f

    def inject_verb(self, verb: str):
        def f(request: fastapi.Request):
            setattr(request.state, 'verb', verb)
        return f

    def is_annotated_dependency(self, value: Any) -> bool:
        if get_origin(value) != Annotated:
            return False
        args = get_args(value)
        if len(args) != 2:
            return False
        return isinstance(args[1], fastapi.params.Depends)

    def is_detail(self) -> bool:
        return self.detail

    def is_model(self, obj: Any) -> bool:
        return obj == self.model

    def is_injectable(self, p: inspect.Parameter):
        return any([
            p.annotation == fastapi.Request,
            p.annotation == IResource,
            p.annotation == APIResourceType,
            p.annotation == ResourceType,
            isinstance(p.default, fastapi.params.Path),
            isinstance(p.default, fastapi.params.Depends),
            self.is_annotated_dependency(p.annotation),
            inspect.isclass(p.annotation) and issubclass(p.annotation, pydantic.BaseModel),
            inspect.isclass(p.annotation) and issubclass(p.annotation, fastapi.Request)
        ])

    def is_namespaced(self) -> bool:
        return self.inspector.is_namespaced(self.model)

    def must_authenticate(self) -> bool:
        return self.authenticated

    def must_authorize(self) -> bool:
        return self.authorized

    def needs_existing(self) -> bool:
        return self.existing

    def on_not_authenticated(self):
        raise HTTPException(
            status_code=401,
            detail="Authentication required."
        )

    def on_unauthorized(self):
        raise HTTPException(
            status_code=403,
            detail=(
                "The request subject is not granted permission "
                "to perform this operation."
            )
        )

    def replaces(self) -> bool:
        return False

    def resource(self, request: fastapi.Request) -> T:
        return getattr(request.state, 'resource')

    async def authorize(
        self,
        ctx: RequestAuthorizationContext,
        verb: RequestVerb
    ) -> None:
        # TODO : The permissions endpoint must be converted to a verb.
        if verb == 'authorize' or not self.must_authorize():
            return
        if not ctx.is_authenticated() and self.must_authenticate():
            self.on_not_authenticated()
        if not ctx.is_authorized():
            self.on_unauthorized()

        if self.permissions and not await ctx.has(self.permissions):
            raise NotImplementedError

    async def inject_existing(
        self,
        request: fastapi.Request,
        resources: ResourceRepository,
        key: RequestObjectReference,
        ifmatch: str | None = fastapi.Header(
            default=None,
            alias='If-Match',
            description=(
                "The `If-Match` header makes a request conditional. If present, "
                "It **must** match the current `.metadata.resourceVersion` value of "
                "the resource. If the values don't match then a `412 Precondition "
                "Failed` response is returned."
            )
        )
    ) -> None:
        if key is None:
            return None
        if ifmatch:
            raise NotImplementedError
        obj = None
        try:
            obj = await self.cache.get(
                key=key.cache_key(self.meta.cache_prefix),
                decoder=self.model
            )
            # Indicate that we retrieved the object from cache
            # so that we know not to cache it when responding.
            if obj is not None:
                setattr(request.state, 'cache_lookup', True)
        except pydantic.ValidationError:
            self.logger.warning(
                "Cache contained an object that violated the current "
                "schema (%s).",
                key
            )
        if obj is not None:
            setattr(request.state, 'resource', obj)
            return
        self.logger.debug("Retrieving resource (%s).", key)
        try:
            obj = await resources.get(key, cast(Any, self.model))
            assert obj is not None
        except resources.DoesNotExist:
            # TODO: Belongs on validator class.
            if self.needs_existing():
                self.fail(404, f'{self.model.__name__} does not exist.')
            obj = None
        setattr(request.state, 'resource', obj)

    async def model_factory(
        self,
        request: fastapi.Request,
        repo: ResourceRepository,
        obj: Any
    ) -> Any:
        assert self.views.create
        if not isinstance(obj, self.views.create):
            raise TypeError(
                f"Can not create a new {self.model.__name__} from "
                f"{type(obj).__name__}"
            )
        # Parse directly from the request JSON content because some
        # fields may be excluded if we would use model_dump().
        try:
            return self.model.model_input(await request.json())
        except pydantic.ValidationError:
            raise
            self.fail(422, f"The object in the request is not a valid {self.model.__name__}.")

    async def on_mutation_request(
        self,
        request: fastapi.Request,
        ctx: IResourceAuthorizationContext,
        repo: IResourceRepository[Any],
        resource: T
    ) -> None:
        """Hook to implement logic prior to mutating a resource."""
        pass

    async def render_to_response(
        self,
        request: fastapi.Request,
        result: T,
        media_type: str | None
    ) -> fastapi.Response:
        headers: dict[str, str] = {}
        if isinstance(result, BaseResource):
            await self.update_cache(request, result) # type: ignore
            headers['ETag'] = result.metadata.resource_version # type: ignore
        if isinstance(result, pydantic.BaseModel):
            result = Response( # type: ignore
                media_type=media_type,
                status_code=self.status_code,
                content=result
            )
        if result is None:
            result = fastapi.Response(status_code=204) # type: ignore
        if not isinstance(result, fastapi.Response):
            raise NotImplementedError
        result.headers.update(headers)
        return result

    async def update_cache(
        self,
        request: fastapi.Request,
        resource: APIResourceType,
        ttl: int | None = None
    ) -> None:
        ttl = ttl or self.ttl
        if ttl is not None and not getattr(request.state, 'cache_lookup', False):
            await self.cache.set(resource.cache_key, resource, ttl=ttl)

    async def __call__(
        self,
        request: fastapi.Request,
        *args: Any, **kwargs: Any
    ) -> fastapi.Response:
        media_type: str | None = getattr(request.state, 'media_type', None)
        repo: ResourceRepository = getattr(request.state, 'resources')
        validator: Validator[Any] | None = getattr(request.state, 'validator', None)
        sig = inspect.signature(self.handle)
        if sig.parameters.get('request'):
            kwargs['request'] = request
        params: dict[str, Any] = {}
        new = None
        for name, value in kwargs.items():
            if name not in sig.parameters:
                continue
            if name in self.input_params:
                value = new = await self.model_factory(request, repo, value)
                if not validator:
                    raise ProgrammingError(
                        "Received input from request but no validator was "
                        f"injected (url: {request.url})"
                    )
                value = await validator.validate(self, new)
            params[name] = value
        try:
            result = await self.handle(**params)
        except Duplicate:
            self.fail(409, "Resource conflict.")
        except (Immutable, ReferentDoesNotExist) as e:
            self.fail(409, e.detail)
        return await self.render_to_response(request, result, media_type=media_type)

    def _get_openapi_responses(self) -> dict[int, Any]:
        responses = self.get_openapi_responses()
        for status_code, response in responses.items():
            headers: dict[str, Any] = response.setdefault('headers', {})
            headers.update(RESPONSE_CONTENT_TYPE)
            if headers.get('Content-Type'):
                headers['Content-Type']['examples'] = [
                    'application/json',
                    'application/yaml'
                ]
            if self.is_detail() and (200 <= status_code < 300):
                headers.update(RESPONSE_ETAG_RESOURCE)

        if not self.authenticated:
            responses.pop(401, None)
            responses.pop(403, None)
        return responses


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetJsonSchemaHandler

from canonical.lib.utils.encoding import b64decode_json
from canonical.lib.utils.encoding import b64encode_json


__all__: list[str] = [
    'Base64JSON'
]


class Base64JSON(dict[str, Any]):

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.str_schema(),
            python_schema=core_schema.union_schema([
                core_schema.chain_schema([
                    core_schema.union_schema([
                        core_schema.bytes_schema(max_length=0),
                        core_schema.str_schema(max_length=0)
                    ]),
                    core_schema.no_info_plain_validator_function(dict)
                ]),
                core_schema.chain_schema([
                    core_schema.union_schema([
                        core_schema.is_instance_schema(bytes),
                        core_schema.is_instance_schema(str),
                    ]),
                    core_schema.no_info_plain_validator_function(b64decode_json),
                    core_schema.dict_schema()
                ]),
                core_schema.dict_schema()
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(cls.serialize),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.str_schema())

    @staticmethod
    def serialize(value: dict[str, Any]) -> bytes:
        if not value:
            return b''
        return b64encode_json(value)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Generic
from typing import Protocol
from typing import TypeVar

from canonical.exceptions import DoesNotExist
from canonical.exceptions import Duplicate
from .itransaction import ITransaction
from .ityped import ITyped


T = TypeVar('T')
M = TypeVar('M')
X = TypeVar('X', contravariant=True)


class IStorage(Protocol, Generic[T, X]):
    """The interface for storage classes. A storage class knows how
    to retrieve and persist objects of a specific type.
    """
    __module__: str = 'canonical.protocols'
    DoesNotExist: type[Exception] = DoesNotExist
    Duplicate: type[Exception] = Duplicate

    async def allocate(self, obj: Any | type[Any]) -> int:
        ...

    async def exists(self, key: Any) -> bool:
        ...

    async def delete(
        self,
        object: M,
        transaction: ITransaction | None = None,
        **kwargs: Any
    ) -> M: ...

    async def first(
        self,
        model: type[M],
        sort: list[str] | None = None,
        transaction: ITransaction | None = None,
        **kwargs: Any
    ) -> M | None:
        ...

    async def get(
        self,
        key: ITyped[M],
        *,
        model: type[M] | None = None,
        cached: bool = False,
        max_age: int = 0,
        transaction: ITransaction | None = None,
        **kwargs: Any
    ) -> M:
        ...

    async def persist(
        self,
        object: T,
        transaction: ITransaction | None = None,
        **kwargs: Any
    ) -> T:
        ...

    async def pop(
        self,
        key: ITyped[M] | int | str,
        model: type[M] | None = None,
        sort: list[str] | None = None,
        transaction: X | None = None,
        **kwargs: Any
    ) -> M:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio

import google.auth
from google.auth.credentials import Credentials


credential: Credentials | None = None
lock = asyncio.Lock()


async def get_credential() -> Credentials:
    global credential
    async with lock:
        if credential is None:
            credential, _ = google.auth.default() # type: ignore
    return credential # type: ignore




# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import ClassVar

import fastapi
import fastapi.params
import pydantic

from canonical.ext.fastapi.params import HTTPClient
from canonical.lib.protocols import ICache
from .clientendpointroutehandler import ClientEndpointRouteHandler
from .endpointrequesthandler import EndpointRequestHandler


class BaseEndpoint:
    cache: ICache
    http: HTTPClient
    http_methods: ClassVar[list[str]] = [
        'GET',
        'HEAD',
        'POST',
        'PUT',
        'PATCH',
        'DELETE',
        'OPTIONS',
        'TRACE',
    ]

    dependencies: list[Any] = []
    model: Any
    name: str | None = None
    path: str
    request: fastapi.Request
    response_model: type[pydantic.BaseModel] | None
    route_class = ClientEndpointRouteHandler
    status_code: int = 200
    summary: str
    tags: list[str] = ['OAuth 2.x/OpenID Connect']

    @classmethod
    def as_endpoint(cls, method: str, **kwargs: Any) -> EndpointRequestHandler:
        return EndpointRequestHandler(
            endpoint=cls,
            handle=getattr(cls, method),
            **kwargs
        )

    @classmethod
    def as_router(
        cls,
        **kwargs: Any
    ) -> fastapi.APIRouter:
        router = fastapi.APIRouter(
            route_class=cls.route_class
        )
        for method in map(str.lower, cls.http_methods):
            if not hasattr(cls, method):
                continue
            router.add_api_route(
                methods=[str.upper(method)],
                path=cls.path,
                name=cls.name,
                endpoint=cls.as_endpoint(method, **kwargs),
                dependencies=cls.dependencies,
                status_code=cls.status_code,
                response_model=cls.response_model,
                response_model_by_alias=False,
                tags=list(cls.tags),
                **cls.get_openapi_params(method)
            )
        return router

    @classmethod
    def get_openapi_params(cls, method: str) -> dict[str, Any]:
        return {
            'summary': cls.summary
        }

    async def setup(self):
        pass

    async def _setup(self):
        await self.setup()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from typing import Any


from canonical.ext.fastapi.mixins import QueryModelMixin
from canonical.ext.oauth.protocols import IRegisteredClient
from canonical.ext.oauth.types import ResponseTypeLiteral
from canonical.ext.oauth.types import RedirectURI
from canonical.ext.oauth.types import URLSafeClientID
from canonical.lib import FormDataModel
from canonical.lib.types import HTTPResourceLocator


class AuthorizationRequest(FormDataModel, QueryModelMixin):
    model_config = {'extra': 'forbid'}
    client_id: URLSafeClientID | str
    response_type: ResponseTypeLiteral
    redirect_uri: RedirectURI | None = None
    state: str | None = None

    @classmethod
    def new(cls, **kwargs: Any):
        kwargs.setdefault('state', secrets.token_hex(32))
        return cls.model_validate(kwargs)

    def authorize(self, authorization_endpoint: HTTPResourceLocator) -> HTTPResourceLocator:
        return authorization_endpoint.with_query(**self.model_dump(mode='json', exclude_none=True))

    def proxy(self, endpoint: str, client_id: str, redirect_uri: str, state: str) -> str:
        """Proxies the authorization request. Return a string containing the
        full authorization URL, with the `client_id`,  `redirect_uri` and
        `state` parameters replaced by the given values.
        """
        clone = self.model_validate({
            **self.model_dump(mode='json'),
            'client_id': client_id,
            'redirect_uri': redirect_uri,
            'state': state
        })
        return str(clone.with_endpoint('GET', endpoint))

    def validate_client(self, client: IRegisteredClient):
        if not client.allows_redirect(self.redirect_uri):
            raise NotImplementedError(f"Redirection to {self.redirect_uri} is not allowed.")
        if not client.allows_response_type(self.response_type):
            raise NotImplementedError(
                "The client does not allow the requested response type."
            )

    def __str__(self):
        return f'{self._url}?{self.model_dump_urlencoded()}'


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
from typing import cast
from typing import overload
from typing import Generic
from typing import TypeVar


T = TypeVar('T', bound=int | str)


class ReferenceDescriptor(Generic[T]):

    def __init__(self, attname: str) -> None:
        self.attname = attname

    @overload
    def __get__(self, obj: None, cls: None) -> ReferenceDescriptor[T]:
        ...

    @overload
    def __get__(self, obj: object, cls: type[object]) -> T:
        ...

    def __get__(
        self, obj: object | None, cls: type[object] | None = None
    ) -> ReferenceDescriptor[T] | T:
        if obj is None:
            return self
        return cast(T, obj.__dict__[self.attname])


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from typing import Any

import fastapi

from canonical.ext.api.bases import BaseResourceBuilder
from canonical.ext.fastapi.params import ResourceRepository
from canonical.ext.oauth.models import AuthorizationRequest
from canonical.ext.oauth.models import ServerMetadata
from canonical.ext.oauth.protocols import IRegisteredClient
from canonical.lib.types import HTTPResourceLocator
from .authorizationstate import AuthorizationState


class AuthorizationStateBuilder(BaseResourceBuilder[AuthorizationState]):
    client: IRegisteredClient | None = None
    mode: str
    model = AuthorizationState

    def __init__(self, request: fastapi.Request, repo: ResourceRepository):
        super().__init__(repo)
        self.request = request
        self.mode = 'proxy'

    def get_model_input(self) -> dict[str, Any]:
        return {
            **super().get_model_input(),
            'status': {
                'changed': self.now,
                'version': 1,
                'message': '',
                'current': 'Created'
            }
        }

    def forward(self, forward_to: str):
        self.mode = self.spec['redirect_mode'] = 'forwarding'
        self.spec['forward_url'] = forward_to

    def proxy(self, callback_uri: str):
        self.mode = self.spec['redirect_mode'] = 'proxy'
        self.spec['proxy_redirection_endpoint'] = callback_uri

    def with_authorization_endpoint(self, url: str):
        self.spec['authorization_endpoint'] = url

    def with_callback(self, redirect_uri: str):
        self.spec['redirect_uri'] = redirect_uri

    def with_client(self, client: IRegisteredClient):
        self.spec['client_id'] = client.id

    def with_client_id(self, client_id: str):
        self.spec['client_id'] = client_id

    def with_issuer(self, issuer: str):
        self.spec['issuer'] = issuer

    def with_metadata(self, metadata: ServerMetadata):
        self.with_issuer(metadata.issuer)
        if metadata.authorization_endpoint:
            self.with_authorization_endpoint(metadata.authorization_endpoint)
        if metadata.token_endpoint:
            self.with_token_endpoint(metadata.token_endpoint)

    def with_params(
        self,
        params: AuthorizationRequest | HTTPResourceLocator,
    ):
        if isinstance(params, HTTPResourceLocator):
            params = AuthorizationRequest.model_validate(params.query)
            self.with_client_id(str(params.client_id))
            if params.state:
                self.metadata['name'] = params.state
        self.spec['params'] = params

    def with_redirect_uri(self, uri: str):
        self.spec['redirect_uri'] = uri

    def with_token_endpoint(self, url: str):
        self.spec['token_endpoint'] = url

    async def prepare(self):
        if not self.metadata.get('name'):
            self.metadata['name'] = secrets.token_hex(32)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .bases import BaseNamespace
from .namespacespec import NamespaceSpec


class Namespace(
    BaseNamespace[str],
    group='',
    version='v1',
    plural='namespaces',
    type='webiam.io/namespace'
):
    model_config = {'populate_by_name': True}

    spec: NamespaceSpec = pydantic.Field(
        default_factory=NamespaceSpec,
        description="Defines the behavior of the `Namespace`."
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Union

import pydantic

from .rsaverificationkey import RSAVerificationKey
from .symmetricencryptionkey import SymmetricEncryptionKey
from .symmetricsigningkey import SymmetricSigningKey


__all__: list[str] = [
    'JSONWebKey'
]


JWKType = Union[
    RSAVerificationKey,
    SymmetricEncryptionKey,
    SymmetricSigningKey
]


class JSONWebKey(pydantic.RootModel[JWKType]):

    @property
    def kid(self):
        return self.root.kid

    @property
    def thumbprint(self):
        return self.root.thumbprint

    def verify(
        self,
        signature: bytes,
        payload: bytes,
        *,
        alg: str | None = None
    ) -> bool:
        # Signature does not verify if the given algorithm does not
        # match the key algorithm.
        if len({alg, self.root.alg}) > 1:
            return False
        return self.root.verify(signature, payload, alg=alg)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .baseencryptionresult import BaseEncryptionResult


class DHEncryptionResult(BaseEncryptionResult):
    epk: dict[str, int | str]

    @property
    def aad(self) -> bytes:
        return b''

    @property
    def iv(self) -> None:
        return None

    @property
    def tag(self) -> None:
        return None

    def __bytes__(self) -> bytes:
        raise NotImplementedError


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
import dataclasses
import hashlib
from typing import Any
from typing import TypeVar
from typing import TYPE_CHECKING

import pydantic

from .basereference import BaseReference
if TYPE_CHECKING:
    from .apimodel import APIModel
    from .apiversionedmeta import APIVersionedMeta
    from .resourcekey import ResourceKey



A = TypeVar('A', bound='APIModel')
T = TypeVar('T')


class APIModelInspector:

    @dataclasses.dataclass
    class ModelStateTuple:
        domain: type[APIModel]
        create: type[pydantic.BaseModel] | None
        update: type[pydantic.BaseModel] | None
        stored: type[pydantic.BaseModel] | None
        output: type[pydantic.BaseModel] | None

    def cache_key(self, model: Any, key: ResourceKey) -> str:
        """Return a string identifying a single resource to be
        used as a cache key.
        """
        meta = self.inspect(model)
        h = hashlib.sha3_256()
        h.update(str.encode(meta.api_version))
        h.update(str.encode(meta.kind))
        raise NotImplementedError
        return h.hexdigest()

    def get_models(self, model: type[APIModel]) -> ModelStateTuple:
        return self.ModelStateTuple(
            model,
            getattr(model, '__create_model__', None),
            None, None, None
            #model.__update_model__,
            #model.__stored_model__,
            #model.__output_model__,
        )

    def get_reference_tree(
        self,
        model: APIModel,
        _depth: int = -1,
        parent: APIModel | None = None
    ) -> list[tuple[int, BaseReference]]:
        """Traverse the object schema to find references."""
        depth = _depth + 1
        refs: list[tuple[int, BaseReference]] = []
        for name in model.model_fields:
            v = getattr(self, name)
            if isinstance(v, BaseReference):
                refs.append((depth, v))
                continue
            if isinstance(v, APIModel):
                refs.extend(self.get_reference_tree(v, depth, model))
        return refs

    def has_state(self, model: type[T]) -> bool:
        return bool(model.model_fields.get('status'))

    def inspect(self, model: type[T]) -> APIVersionedMeta[Any]:
        try:
            return getattr(model, '__meta__')
        except AttributeError:
            raise TypeError(
                f"{model.__name__} is not a known type that can be described "
                f"by {type(self).__name__}."
            )

    def is_dirty(self, instance: APIModel) -> bool:
        """Return a boolean if the data maintained in an instance
        must be written back to its primary or origin storage.
        """
        return instance._dirty # type: ignore

    def is_namespaced(self, model: type[APIModel]):
        """Return a boolean indicating if the model is namespaced."""
        return model.__namespaced__


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging

from canonical.ext.cache import MemoryCache
from canonical.ext.httpx import AsyncClient
from canonical.ext.jose.models import JSONWebKeySet
from canonical.ext.jose.protocols import ITokenSignatureVerifier
from canonical.ext.jose.types import JOSEHeaderDict
from canonical.ext.jose.types import JWTDict
from canonical.lib.protocols import ICache


class ServiceAccountTokenVerifier(ITokenSignatureVerifier):
    cache_prefix: str = 'google:serviceaccount:jwks'
    cache_ttl: int = 3600
    jwks_endpoint: str = 'https://www.googleapis.com/service_accounts/v1/metadata/jwk'
    logger: logging.Logger = logging.getLogger('uvicorn')

    def __init__(
        self,
        http: AsyncClient,
        cache: ICache = MemoryCache(scope='global')
    ):
        self.cache = cache
        self.http = http

    def cache_key(self, email: str) -> str:
        return str.join(':', [self.cache_prefix, email])

    def is_selectable(self, header: JOSEHeaderDict, claims: JWTDict) -> bool:
        # When obtaining the token, the caller must set the `iss` claim to
        # the service account email. The `kid` header claim is assumed to
        # be always set by Google.
        return all([
            str.endswith(claims.get('iss') or '', '.iam.gserviceaccount.com'),
            'kid' in header,
            'alg' in header
        ])

    async def verify_token(
        self,
        *,
        header: JOSEHeaderDict,
        claims: JWTDict,
        signature: bytes,
        payload: bytes
    ) -> bool:
        if not self.is_selectable(header, claims):
            return False
        assert 'alg' in header
        assert 'kid' in header
        assert 'iss' in claims
        jwks = await self.get_service_account_jwks(claims['iss'])
        if not jwks.has(kid=header['kid']):
            return False
        return jwks.verify(signature, payload, kid=header['kid'], alg=header['alg'])

    async def get_service_account_jwks(self, email: str) -> JSONWebKeySet:
        jwks = await self.cache.get(self.cache_key(email), JSONWebKeySet)
        if jwks is None:
            response = await self.http.request(
                method='GET',
                url=f'{self.jwks_endpoint}/{email}',
                response_model=JSONWebKeySet,
                follow_redirects=True
            )
            jwks = JSONWebKeySet()
            if response.status_code >= 400:
                self.logger.warning(
                    "Unable to retrieve Google Service Account JWKS "
                    "(email: %s, status: %s url: %s)",
                    email, response.status_code, response.url
                )
            else:
                jwks = response.result
        return jwks


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import copy
import os

import jinja2
import pydantic
import yaml


class ConfigFileModel(pydantic.BaseModel):

    @classmethod
    def open(cls, fn: str):
        with open(fn) as f:
            t = jinja2.Template(
                f.read(),
                variable_start_string='${',
                variable_end_string='}',
                undefined=jinja2.StrictUndefined
            )
        data = yaml.safe_load(t.render(env=copy.copy(os.environ))) # type: ignore
        return cls.model_validate(data)


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .apidescriptor import APIDescriptor
from .apigroupversionlist import APIGroupVersionList
from .apiresource import APIResource as APIResourceModel
from .apiresourcelist import APIResource
from .apiresourcelist import APIResourceList
from .conditiontype import ConditionType
from .error import Error
from .iresourcequery import IResourceQuery
from .iresourcerepository import IResourceRepository
from .listbase import ListBase
from .localreference import LocalReference
from .namespacedobjectmeta import NamespacedObjectMeta
from .objectmeta import ObjectMeta
from .primarykey import PrimaryKey
from .resource import M as ObjectMetaType
from .resource import Resource
from .resourceinspector import ResourceInspector
from .resourcemeta import ResourceMeta
from .resourceserverlist import ResourceServerList
from .resourcespec import ResourceSpec
from .resourcestatus import ResourceStatus
from .rootresource import ResourceType
from .rootresource import ResourceTypeVar
from .rootresource import RootResource
from .statefulresource import StatefulResource
from .transientmeta import TransientMeta

# Public interface
from canonical.ext.api import LocalObjectReference
from canonical.ext.api import ObjectReference
from .apiversioned import APIVersioned
from .basereference import BaseReference
from .baseresourcerepository import BaseResourceRepository
from .resourcekey import ResourceKey
from .ownerreference import OwnerReference
from .typedobjectreference import TypedObjectReference



__all__: list[str] = [
    'APIDescriptor',
    'APIGroupVersionList',
    'APIResource',
    'APIResourceList',
    'APIVersioned',
    'BaseReference',
    'BaseResourceRepository',
    'ConditionType',
    'Error',
    'Inspectable',
    'IResourceQuery',
    'IResourceRepository',
    'ListBase',
    'LocalObjectReference',
    'LocalReference',
    'Namespace',
    'NamespacedObjectMeta',
    'ObjectMeta',
    'ObjectMetaType',
    'ObjectReference',
    'OwnerReference',
    'PrimaryKey',
    'Resource',
    'ResourceInspector',
    'ResourceKey',
    'ResourceMeta',
    'ResourceServerList',
    'ResourceSpec',
    'ResourceStatus',
    'ResourceType',
    'ResourceTypeVar',
    'RootResource',
    'StatefulResource',
    'TransientMeta',
    'TypedObjectReference',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import typing
from typing import NotRequired
from typing import TypedDict

from httpx import AsyncBaseTransport
from httpx._config import Limits
from httpx._types import AuthTypes
from httpx._types import CertTypes
from httpx._types import CookieTypes
from httpx._types import HeaderTypes
from httpx._types import ProxiesTypes
from httpx._types import ProxyTypes
from httpx._types import QueryParamTypes
from httpx._types import TimeoutTypes
from httpx._types import VerifyTypes
from httpx._types import URLTypes


class AsyncClientParams(TypedDict):
    auth: NotRequired[AuthTypes]
    params: NotRequired[QueryParamTypes]
    headers: NotRequired[HeaderTypes]
    cookies: NotRequired[CookieTypes]
    verify: NotRequired[VerifyTypes]
    cert: NotRequired[CertTypes]
    http1: NotRequired[bool]
    http2: NotRequired[bool]
    proxy: NotRequired[ProxyTypes]
    proxies: NotRequired[ProxiesTypes]
    mounts: NotRequired[
        typing.Mapping[str, NotRequired[AsyncBaseTransport]]
    ]
    timeout: NotRequired[TimeoutTypes]
    follow_redirects: NotRequired[bool]
    limits: NotRequired[Limits]
    max_redirects: NotRequired[int]
    event_hooks: NotRequired[
        typing.Mapping[str, typing.List[typing.Callable[..., typing.Any]]]
    ]
    base_url: NotRequired[URLTypes]
    transport: NotRequired[AsyncBaseTransport]
    app: NotRequired[typing.Callable[..., typing.Any]]
    trust_env: NotRequired[bool]
    default_encoding: NotRequired[typing.Union[str, typing.Callable[[bytes], str]]]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import cast
from typing import overload
from typing import Any
from typing import Literal
from typing import NoReturn
from typing import TypeVar
from typing import Unpack

import httpx
import pydantic
import yaml
from fastapi.exceptions import HTTPException

from canonical.ext.cache import NullCache
from canonical.ext.iam import PermissionQuery
from canonical.ext.iam import PermissionSet
from canonical.ext.iam.types import Permission
from canonical.ext.resource import APIDescriptor
from canonical.ext.resource import APIResource
from canonical.ext.resource import APIResourceList
from canonical.ext.resource import ResourceType
from canonical.ext.httpx import AsyncClient
from canonical.ext.httpx import AsyncClientParams
from canonical.ext.resource import Error
from canonical.lib.protocols import ICache
from canonical.utils import deephash


R = TypeVar('R', bound=ResourceType)
T = TypeVar('T', bound=pydantic.BaseModel)


class ResourceClient(AsyncClient):
    _supported_resources: dict[str, APIResourceList] = {}
    logger: logging.Logger = logging.getLogger('uvicorn')
    mode: Literal['cli', 'impersonated']
    authorization_ttl: int = 600
    permission_ttl: int = 900

    def __init__(
        self,
        mode: Literal['cli', 'impersonated'] = 'impersonated',
        cache: ICache = NullCache(),
        **kwargs: Unpack[AsyncClientParams]
    ):
        headers: dict[str, str] = cast(dict[str, str], kwargs.setdefault('headers', {}))
        headers.setdefault('Accept', 'application/json')
        self.cache = cache
        self.mode = mode
        super().__init__(**kwargs)

    def can_retry_timeout(self, request: httpx.Request):
        return super().can_retry_timeout(request) or any([
            str.endswith(request.url.path, ':permissions')
        ])

    def construct_uri(
        self,
        group: str,
        kind: str,
        name: str | None = None,
        namespace: str | None = None,
        version: str | None = None
    ):
        version = version or 'v1'
        url = f'{version}/{kind}'
        if group:
            url = f'{group}/{url}'
        if namespace is not None:
            url = f'/{group}/{version}/namespaces/{namespace}/{kind}'
        if name is not None:
            url = f'{url}/{name}'
        return url

    def is_discovered(self, api_version: str) -> bool:
        return api_version in self._supported_resources

    def raise_for_status(self, response: httpx.Response) -> NoReturn | None:
        if response.status_code >= 400:
            if response.request.method == 'HEAD':
                if self.mode == 'cli':
                    raise SystemExit(f"Server responded with {response.status_code} (url: {response.url}).")
                raise HTTPException(
                    status_code=response.status_code,
                    detail="No details provided by the server because the upstream method was HEAD."
                )
            ct = response.headers.get('Content-Type') or 'application/json'
            if not str.startswith(ct,'application/json'):
                self.logger.warning(
                    "Unexpected response encoding (expected: %s, actual: %s)",
                    "application/json",
                    ct
                )
            try:
                if ct == 'application/yaml':
                    error = Error.model_validate(yaml.safe_load(response.content)) # type: ignore
                else:
                    error = Error.model_validate_json(response.content)
            except pydantic.ValidationError:
                self.logger.critical(
                    "Unexpected response from server: %s",
                    response.content.decode()
                )
                raise
            status_code = getattr(error.data, 'status_code', 500)
            if self.mode == 'cli':
                raise SystemExit('---\n' + error.model_dump_yaml(indent=2))
            raise HTTPException(
                status_code=status_code,
                detail=error.data.detail
            )

    @overload
    def resource_factory(self, model: None, response: httpx.Response) -> dict[str, Any]:
        ...

    @overload
    def resource_factory(self, model: type[T], response: httpx.Response) -> T:
        ...

    def resource_factory(self, model: type[T] | None, response: httpx.Response) -> T | dict[str, Any]:
        return (
            response.json()
            if not model else
            model.model_validate_json(response.content)
        )

    async def apply(self, obj: dict[str, Any]) -> None:
        """Applies the given specification to the system."""
        metadata: dict[str, Any] | Any = obj.get('metadata')
        if metadata is None or not isinstance(metadata, dict):
            raise TypeError(
                "The metadata field is required and must be a "
                "dictionary."
            )
        name = metadata.get('name')
        generate = bool(metadata.get('generateName'))
        ns = metadata.get('namespace')
        if (name is None or not isinstance(name, (str, int))) and not generate:
            raise TypeError(
                "The .metadata.name field is required and must be a "
                "string."
            )
        resource, api_version = await self.discover_object(obj)
        if (ns is None or not isinstance(ns, str)) and resource.namespaced:
            raise TypeError(
                "The metadata.namespace field is required and must be a "
                "string."
            )
        method = self.create
        if name:
            method = self.replace
            exists = await self.exists(
                resource.path(
                    api_version=api_version,
                    name=name,
                    namespace=ns
                )
            )
            if not exists:
                method = self.create
            result = await method(resource, api_version, ns, name, obj)
        else:
            result = await method(resource, api_version, ns, name, obj)
        return result

    async def create(
        self,
        resource: APIResource,
        api_version: str,
        namespace: str | None,
        name: str | int | None,
        obj: dict[str, Any]
    ) -> None:
        response = await self.post(
            url=resource.path(api_version, namespace=namespace),
            json=obj
        )
        self.raise_for_status(response)

    async def discover_server(self, server: str) -> APIDescriptor:
        response = await self.get(url=server)
        self.raise_for_status(response)
        return APIDescriptor.model_validate_json(response.content)

    async def discover_object(self, obj: dict[str, Any]):
        api_version: str | None = obj.get('apiVersion')
        if api_version is None:
            raise TypeError("The apiVersion field is required.")
        kind = obj.get('kind')
        if kind is None:
            raise TypeError("The kind field is required.")
        if not self.is_discovered(api_version):
            response = await self.get(
                url=api_version,
                headers={'Accept': 'application/json'}
            )
            self.raise_for_status(response)
            self._supported_resources[api_version] = APIResourceList\
                .model_validate_json(response.content)

        return self._supported_resources[api_version].get(kind), api_version

    async def exists(self, url: str) -> bool:
        response = await self.head(
            url=url,
            follow_redirects=True
        )
        if response.status_code not in {200, 404}:
            self.raise_for_status(response)
        return response.status_code == 200

    async def permissions(
        self,
        namespace: str | None,
        permission: str,
        *permissions: str,
        cache_key: str | None = None,
        **kwargs: Any
    ) -> set[Permission]:
        """Retrieve the permissions that the client has on the given
        resource.
        """
        url = f"{self.construct_uri('', 'namespaces', namespace)}:permissions"
        granted: set[Permission] | None = None
        if cache_key:
            cache_key = f'permissions:{cache_key}:{namespace}:{deephash({permission, *permissions}, encode="hex")}'
            granted = await self.cache.get(
                key=cache_key,
                decoder=set[Permission]
            )
        if granted is None:
            self.logger.warning("Retrieving permissions from server (url: %s).", url)
            query = PermissionQuery.factory({
                'permissions': {permission, *permissions}
            })
            headers: dict[str, str] = kwargs.setdefault('headers', {})
            headers.setdefault('Accept', 'application/json')
            response = await self.post(
                headers=headers,
                url=url,
                timeout=None,
                json=query.model_dump(by_alias=True, mode='json')
            )
            if response.status_code >= 400:
                self.logger.warning(
                    "Permissions endpoint responded with %s (endpoint: %s)",
                    response.status_code, response.url
                )
            else:
                result = PermissionSet.model_validate_json(response.content)
                granted = {Permission(x) for x in result.granted}
                if cache_key:
                    await self.cache.set(
                        key=cache_key,
                        value=granted,
                        encoder=set[Permission],
                        ttl=self.permission_ttl
                    )
        return granted or set()

    @overload
    async def request_resource(self, model: None, **kwargs: Any) -> dict[str, Any]:
        ...

    @overload
    async def request_resource(self, model: type[T], **kwargs: Any) -> T:
        ...

    async def request_resource(self, model: type[T] | None, method: str, **kwargs: Any) -> T | dict[str, Any]:
        headers = kwargs.setdefault('headers', {})
        headers['Accept'] = 'application/json'
        response = await self.request(method=method, **kwargs)
        self.raise_for_status(response)
        resource = response.json()
        if model is not None:
            resource = self.resource_factory(model, response)
        return resource

    async def retrieve(
        self,
        group: str,
        kind: str,
        name: str,
        model: type[T] | None = None,
        namespace: str | None = None,
        version: str | None = None,
    ) -> dict[str, Any] | T:
        return await self.request_resource(
            model=model,
            method='GET',
            url=self.construct_uri(group, kind, name, namespace, version)
        )

    async def replace(
        self,
        resource: APIResource,
        api_version: str,
        namespace: str | None,
        name: str | int,
        obj: dict[str, Any]
    ) -> None:
        response = await self.put(
            url=resource.path(api_version, name=name, namespace=namespace),
            headers={'Accept': 'application/json'},
            json=obj,
        )
        if response.status_code == 405:
            self.logger.warning(
                "Object replacement is not supported for %s resources (apiVersion: %s)",
                obj['kind'], obj['apiVersion']
            )
            return
        self.raise_for_status(response)


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar


__all__: list[str] = [
    'KeyFactory'
]

M = TypeVar('M')


class Key(Generic[M]):

    def __iter__(self):
        return iter((None, None))


class KeyFactory(Generic[M]):
    model: type[M]

    def __init__(self, model: type[M]):
        self.model = model


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from .localobjectreference import LocalObjectReference


__all__: list[str] = [
    'ClusterObjectReference'
]


class ClusterObjectReference(LocalObjectReference):
    """A `ClusterlObjectReference` instance contains enough
    information to let you locate the referenced object
    inside the same cluster.
    """

    @property
    def model(self) -> type[Any]:
        assert self._reference_model is not None
        return self._reference_model

    def as_name(self) -> str:
        return self.name

    def attach_to_namespace(self, namespace: str) -> None:
        raise NotImplementedError

    def get_namespace(self) -> str | None:
        return None

    def in_namespace(self, namespace: str) -> bool:
        return False

    def is_cluster(self) -> bool:
        return True

    def is_namespaced(self) -> bool:
        return True


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from collections import OrderedDict
from typing import get_origin
from typing import Any
from typing import Annotated
from typing import Callable

import fastapi
import fastapi.params

from canonical.exceptions import ProgrammingError


__all__: list[str] = [
    'request_state'
]


UNSET = object()


def request_state(
    name: str,
    annotation: Any,
    required: bool = True,
    default: object = UNSET
) -> Any:
    if not bool(default != UNSET) ^ bool(required):
        raise TypeError(
            "The `required` and `default` parameters are mutually "
            "exclusive."
        )

    def f(request: fastapi.Request):
        v = getattr(request, name, default)
        if v == UNSET and required:
            raise ProgrammingError(
                f"Dependency requires request.state.{name} "
                "to be present."
            )
        setattr(request.state, name, v)
        return v

    return fastapi.Depends(f)


def inject_state(name: str, required: bool = True) -> Any:
    def f(request: fastapi.Request):
        try:
            return getattr(request.state, name)
        except AttributeError:
            if required:
                raise ProgrammingError(
                    f"Dependency {name} is requested from the request state "
                    "but it was not injected."
                )
            return None
    return fastapi.Depends(f)


def update_signature(f: Callable[..., Any], *args: Any, **kwargs: Any):
    sig = inspect.signature(f)
    params: OrderedDict[str, inspect.Parameter] = OrderedDict()
    if 'request' in sig.parameters:
        params['request'] = sig.parameters['request']
    for i, dependency in enumerate(args):
        name = f'__arg_{i}'
        if isinstance(dependency, fastapi.params.Depends):
            kwargs[name] = dependency
            continue
        if not get_origin(dependency) == Annotated:
            continue
        params[name] = inspect.Parameter(
            kind=inspect.Parameter.POSITIONAL_ONLY,
            name=name,
            annotation=dependency
        )
        
    for name, dependency in kwargs.items():
        if not isinstance(dependency, fastapi.params.Depends):
            if not callable(dependency):
                dependency = lambda: dependency
            dependency = fastapi.params.Depends(dependency)
        params[name] = inspect.Parameter(
            kind=inspect.Parameter.KEYWORD_ONLY,
            name=name,
            default=dependency,
            annotation=Any
        )
    sig = sig.replace(parameters=list(params.values()))
    return setattr(f, '__signature__', sig)


def inject(*args: Any, **kwargs: Any):
    def f(request: fastapi.Request, **kwargs: Any):
        for k, v in kwargs.items():
            setattr(request.state, k, v)

    update_signature(f, *args, **kwargs)
    return [fastapi.Depends(f)]


def requires(
    dependant: Callable[..., Any],
    *args: Any,
    **kwargs: Any,
) -> Any:
    def f(**params: Any):
        args: list[Any] = []
        kwargs: dict[str, Any] = {}
        for k, v in sorted(params.items(), key=lambda x: x[0]):
            if k.startswith('__arg'):
                args.append(v)
                continue
            kwargs[k] = v
        return dependant(*args, **kwargs)

    update_signature(f, *args, **kwargs)
    return fastapi.Depends(f)


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Literal
from typing import NotRequired
from typing import TypedDict
from typing import Unpack

from pydantic import Field
from pydantic.fields import AliasChoices
from pydantic.fields import AliasPath
from pydantic.fields import JsonDict
from pydantic.fields import PydanticUndefined
from pydantic.fields import PydanticUndefined
from pydantic.types import Discriminator

from .apimodelfieldinfo import APIModelFieldInfo


_Unset: Any = PydanticUndefined


class APIModelFieldParams(TypedDict):
    default_factory: NotRequired[Callable[[], Any] | None]
    alias: NotRequired[str | None]
    alias_priority: NotRequired[int | None]
    validation_alias: NotRequired[str | AliasPath | AliasChoices | None]
    serialization_alias: NotRequired[str | None]
    title: NotRequired[str | None]
    description: NotRequired[str | None]
    examples: NotRequired[list[Any] | None]
    exclude: NotRequired[bool | None]
    discriminator: NotRequired[str | Discriminator | None]
    json_schema_extra: NotRequired[JsonDict | Callable[[JsonDict], None] | None]
    frozen: NotRequired[bool | None]
    validate_default: NotRequired[bool | None]
    repr: NotRequired[bool]
    init_var: NotRequired[bool | None]
    kw_only: NotRequired[bool | None]
    pattern: NotRequired[str | None]
    strict: NotRequired[bool | None]
    gt: NotRequired[float | None]
    ge: NotRequired[float | None]
    lt: NotRequired[float | None]
    le: NotRequired[float | None]
    multiple_of: NotRequired[float | None]
    allow_inf_nan: NotRequired[bool | None]
    max_digits: NotRequired[int | None]
    decimal_places: NotRequired[int | None]
    min_length: NotRequired[int | None]
    max_length: NotRequired[int | None]
    union_mode: NotRequired[Literal['smart', 'left_to_right']]


def APIModelField(
    default: Any = PydanticUndefined,
    *,
    when: set[Literal['create', 'update', 'store', 'view']] | None = _Unset,
    encrypt: bool = False,
    **kwargs: Unpack[APIModelFieldParams]
) -> Any:
    return APIModelFieldInfo.from_field_info(
        field=Field(default=default, **kwargs),
        encrypt=encrypt,
        when=when
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .canonicalexception import CanonicalException


class ReferentDoesNotExist(CanonicalException, LookupError):
    """An exception class that indicates that a certain referent
    does not exist.
    """
    __module__: str = 'canonical.exceptions'

    def __init__(self, detail: str):
        self.detail = detail


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from canonical.ext.fastapi import ResourceRouter
from canonical.ext.fastapi import params
from canonical.ext.fastapi import verbs

from canonical.ext.oauth.params import RedirectedState
from ..clusterclient import ClusterClient
from ..clusterprovider import ClusterProvider
from ..provider import Provider
from ..resources import AuthorizationState


__all__: list[str] = [
    'authorizationstates',
    'clusterclients',
    'clusterproviders',
    'providers',
]

authorizationstates = ResourceRouter.new(AuthorizationState)
authorizationstates.add_verb(verbs.Retrieve)
client = fastapi.APIRouter(prefix='/oauth/v2')
clusterclients = ResourceRouter.new(ClusterClient) 
clusterclients.add_verb(verbs.DefaultCreate)
clusterclients.add_verb(verbs.DefaultReplace)
clusterclients.add_verb(verbs.Retrieve, ttl=3600)
clusterproviders = ResourceRouter.new(ClusterProvider)
clusterproviders.add_verb(verbs.DefaultCreate)
providers = ResourceRouter.new(Provider)
providers.add_verb(verbs.DefaultCreate)


@clusterproviders.verb(verbs.Retrieve, authenticated=False, ttl=3600)
async def retrieve_cluster(
    verb: verbs.Retrieve[ClusterProvider],
    client: params.HTTPClient,
    obj: ClusterProvider
):
    if obj.is_discoverable():
        await obj.discover(client)
    return obj


@client.get('/callback', name='oauth2.callback')
def f(
    request: fastapi.Request,
    state: RedirectedState
):
    if not state.is_proxy() or state.is_failed():
        raise NotImplementedError
    return fastapi.responses.RedirectResponse(
        status_code=302,
        url=state.get_return_url(
            iss=f'{request.url.scheme}://{request.url.netloc}'
        )
    )
    


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import TypeVar

from pydantic.main import BaseModel

from canonical.ext.resource import Error
from .default import Default


T = TypeVar('T')


class Proxy(Default[T]):
    detail: bool = True
    existing: bool = True
    include_in_schema: bool = False

    def get_endpoint_summary(self) -> str:
        return "Proxy request to a remote service"

    def get_response_description(self) -> str:
        return "Response from the remote service"

    def get_response_model(self) -> type[BaseModel] | None:
        return None

    def get_openapi_responses(self) -> dict[int, dict[str, Any]]:
        return {
            511: {
                'model': Error,
                'description': {
                    'Authentication is required to obtain access to the '
                    'upstream service.'
                }
            }
        }

    def get_request_methods(self) -> list[str]:
        return [
            'OPTIONS',
            'HEAD',
            'GET',
            'POST',
            'PUT',
            'DELETE',
            'PATCH',
            # Not supported by FastAPI
            # 'TRACE',
        ]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

from canonical.ext.jose.types import JOSEHeaderDict
from canonical.ext.jose.types import JWTDict


class ITokenSignatureVerifier(Protocol):
    __module__: str = 'canonical.ext.jose'

    def is_selectable(self, header: JOSEHeaderDict, claims: JWTDict) -> bool:
        """Return a boolean indicating if the verifier is selectable,
        based on the claim in a JSON Web Token (JWT). The default
        implementation always returns ``True``.
        """
        return True

    async def verify_token(
        self,
        *,
        header: JOSEHeaderDict,
        claims: JWTDict,
        signature: bytes,
        payload: bytes,
    ) -> bool:
        ...



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi

from canonical.ext.fastapi.params import ResourceRepository
from canonical.ext.oauth.bases import AuthorizationServerEndpoint
from canonical.ext.oauth.params import AuthorizationRequest
from canonical.ext.oauth.types import URLSafeClientID
from canonical.ext.oauth.protocols import IRegisteredClient
from canonical.ext.oauth.clusterclient import ClusterClient
from .useragentroutehandler import UserAgentRouteHandler


class AuthorizeEndpoint(AuthorizationServerEndpoint):
    cache_ttl: int = 3600
    client_resource: ClusterClient
    name = 'oauth2.authorize'
    params: AuthorizationRequest
    path = '/authorize'
    route_class = UserAgentRouteHandler
    response_model = None
    resources: ResourceRepository
    status_code = 302
    summary = 'Authorize Endpoint'

    @property
    def client(self) -> IRegisteredClient:
        assert self._client is not None
        return self._client

    def get_client_meta(self) -> tuple[URLSafeClientID, str, type]:
        assert isinstance(self.params.client_id, URLSafeClientID)
        return (
            self.params.client_id,
            self.params.client_id.cache_key,
            self.params.client_id.model
        )

    async def discover(self, client: Any):
        pass

    async def get(self) -> fastapi.Response:
        raise NotImplementedError

    async def setup(self):
        await super().setup()
        await self.discover(self.client)
        await self.validate_client(self.client, self.params)

    async def validate_client(self, client: IRegisteredClient, params: AuthorizationRequest):
        params.validate_client(client)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Iterable

from .mediatype import get_best_match


class MediaTypeSelector:

    def __init__(self, allow: Iterable[str]):
        self.allow = set(allow)

    def select(self, header: str | None):
        if header is None:
            return None
        return get_best_match(header, self.allow)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
import datetime
from typing import Any
from typing import TypeVar
from typing import TYPE_CHECKING

import pydantic

from canonical.ext.builder import BaseBuilder
from .resourcerepository import BaseResourceRepository
if TYPE_CHECKING:
    from .. import APIResourceType


T = TypeVar('T', bound=pydantic.BaseModel)


class BaseResourceBuilder(BaseBuilder[T]):
    model: type[T]
    metadata: dict[str, Any]
    block_owner_deletion: bool = True

    def __init__(
        self,
        repo: BaseResourceRepository,
        initial: dict[str, Any] | None = None
    ):
        super().__init__()
        self.now = datetime.datetime.now(datetime.timezone.utc)
        self.metadata = {
            'annotations': {},
            'created': self.now,
            'updated': self.now,
            'labels': {},
            'owner_references': [],
            'uid': None
        }
        self.repo = repo
        self.spec = initial or {}

    def annotate(self, name: str, value: str):
        self.metadata['annotations'][name] = value

    def get_model_input(self) -> dict[str, Any]:
        return {
            'metadata': self.metadata,
            'spec': self.spec
        }

    def has_controller(self) -> bool:
        return any([x.controller for x in self.metadata['owner_references']])

    def label(self, name: str, value: str):
        self.metadata['labels'][name] = value

    def own(self, owner: APIResourceType, controller: bool = False):
        from ..ownerreference import OwnerReference # TODO
        assert owner.metadata.uid
        meta = owner.__meta__
        if controller and meta.is_namespaced():
            self.metadata['namespace'] = owner.get_namespace()
        self.metadata['owner_references'].append(
            OwnerReference.model_validate({
                'api_version': owner.api_version,
                'kind': owner.kind,
                'uid': owner.metadata.uid,
                'name': owner.metadata.name,
                'controller': controller,
                'block_owner_deletion': self.block_owner_deletion
            })
        )

    async def prepare(self):
        pass

    async def build(self) -> T:
        return self.model.model_validate(self.get_model_input())

    async def _build(self):
        self.metadata['uid'] = await self.repo.allocate(self.model)
        await self.prepare()
        return await self.build()

    def __await__(self):
        return self._build().__await__()